import sys
import os
import inspect
import collections
import functools
import argparse
import re
import pprint
from enum import Enum, EnumMeta
import time
from contextlib import contextmanager
from threading import Thread, Lock


def report_exitcode(code, name='process'):
    """
    Given exit code of a process, throw error / print result as desired
    """
    if code == 0:
        print('[Warning] {} exited with exit code 0'.format(name))
    elif code > 0:
        raise ValueError('[Warning] {} exited with exit code {}'.format(name, code))
    else:
        raise ValueError('[Warning] {} terminated by signal {}'.format(name, code))


def _get_qualified_type_name(type_):
    name = str(type_)
    r = re.compile("<class '(.*)'>")
    match = r.match(name)
    if match:
        return match.group(1)
    else:
        return name


def assert_type(x, expected_type, message=''):
    assert isinstance(x, expected_type), (
        (message + ': ' if message else '')
        + 'expected type `{}`, actual type `{}`'.format(
            _get_qualified_type_name(expected_type),
            _get_qualified_type_name(type(x))
        )
    )
    return True


def print_(*objs, h='', **kwargs):
    """
    Args:
      *objs: objects to be pretty-printed
      h: header string
      **kwargs: other kwargs to pass on to ``pprint()``
    """
    if h:
        print('=' * 20, h, '=' * 20)
    for obj in objs:
        print(pprint.pformat(obj, indent=4, **kwargs))
    if h:
        print('=' * (42 + len(h)))


class _GetItemEnumMeta(EnumMeta):
    """
    Hijack the __getitem__ method from metaclass, because subclass cannot
        override magic methods. More informative error message.
    """
    def __getitem__(self, option):
        enum_class = None
        for v in self.__members__.values():
            enum_class = v.__class__
            break
        assert enum_class is not None, \
            'must have at least one option in StringEnum'
        return get_enum(enum_class, option)


class StringEnum(Enum, metaclass=_GetItemEnumMeta):
    """
    https://docs.python.org/3.4/library/enum.html#duplicatefreeenum
    The created options will automatically have the same string value as name.

    Support [] subscript, i.e. MyFruit['orange'] -> MyFruit.orange
    """
    def __init__(self, *args, **kwargs):
        self._value_ = self.name


def create_string_enum(class_name, option_names):
    assert_type(option_names, str)
    assert_type(option_names, list)
    return StringEnum(class_name, option_names)


def get_enum(enum_class, option):
    """
    Args:
        enum_class:
        option: if the value doesn't belong to Enum, throw error.
            Can be either the str name or the actual enum value
    """
    assert issubclass(enum_class, StringEnum)
    if isinstance(option, enum_class):
        return option
    else:
        assert_type(option, str)
        option = option.lower()
        options = enum_class.__members__
        if option not in options:
            raise ValueError('"{}" is not a valid option for {}. '
                             'Available options are {}.'
             .format(option, enum_class.__name__, list(options)))
        return options[option]


def fformat(float_num, precision):
    """
    https://stackoverflow.com/a/44702621/3453033
    """
    assert isinstance(precision, int) and precision > 0
    return ('{{:.{}f}}'
            .format(precision)
            .format(float_num)
            .rstrip('0')
            .rstrip('.'))


def mean(numbers):
    return float(sum(numbers)) / max(len(numbers), 1)


def is_sequence(obj):
    """
    Returns:
      True if the sequence is a collections.Sequence and not a string.
    """
    return (isinstance(obj, collections.Sequence)
            and not isinstance(obj, str))


def include_keys(include, d):
    """
    Pick out the `include` keys from a dict

    Args:
      include: list or set of keys to be included
      d: raw dict that might have irrelevant keys
    """
    assert is_sequence(include)
    return {k: v for k, v in d.items() if k in set(include)}


def exclude_keys(exclude, d):
    """
    Remove the `exclude` keys from a kwargs dict.

    Args:
      exclude: list or set of keys to be excluded
      d: raw dict that might have irrelevant keys
    """
    assert is_sequence(exclude)
    return {k: v for k, v in d.items() if k not in set(exclude)}


def iter_last(iterable):
    """
    For processing the last element differently
    Yields: (is_last=bool, element)
    """
    length = len(iterable)
    return ((i == length-1, x) for i, x in enumerate(iterable))


def _get_bound_args(func, *args, **kwargs):
    """
    https://docs.python.org/3/library/inspect.html#inspect.BoundArguments
    def f(a, b, c=5, d=6): pass
    get_bound_args(f, 3, 6, d=100) -> {'a':3, 'b':6, 'c':5, 'd':100}

    Returns:
        OrderedDict of bound arguments
    """
    arginfo = inspect.signature(func).bind(*args, **kwargs)
    arginfo.apply_defaults()
    return arginfo.arguments


class Timer(object):
    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.interval = self.end - self.start
        print('elapsed', self.interval)


class SaveInitArgsMeta(type):
    """
    Bounded arguments:
    https://docs.python.org/3/library/inspect.html#inspect.BoundArguments

    Store the captured constructor arguments to <instance>._init_args
    as OrderedDict. Can be retrieved by the property method <obj>.init_args
    """
    def __init__(cls, name, bases, attrs):
        # WARNING: must add class method AFTER super.__init__
        # adding attrs['new-method'] before __init__ has no effect!
        super().__init__(name, bases, attrs)
        @property
        def init_args(self):
            return self._init_args
        cls.init_args = init_args

    def __call__(cls, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        obj._init_args = _get_bound_args(obj.__init__, *args, **kwargs)
        return obj


class SaveInitArgs(metaclass=SaveInitArgsMeta):
    """
    Either use metaclass hook:
        class MyObj(metaclass=SaveInitArgsMeta)
    or simply inherit
        class MyObj(SaveInitArgs)
    """
    pass


class AutoInitializeMeta(type):
    """
    Call the special method ._initialize() after __init__.
    Useful if some logic must be run after the object is constructed.
    For example, the following code doesn't work because `self.y` does not exist
    when super class calls self._initialize()

    class BaseClass():
        def __init__(self):
            self._initialize()

        def _initialize():
            self.x = self.get_x()

        def get_x(self):
            # abstract method that only subclass

    class SubClass(BaseClass):
        def __init__(self, y):
            super().__init__()
            self.y = y

        def get_x(self):
            return self.y * 3

    Fix:
    class BaseClass(metaclass=AutoInitializeMeta):
        def __init__(self):
            pass
            # self._initialize() is now automatically called after __init__

        def _initialize():
            print('INIT', self.x)

        def get_x(self):
            # abstract method that only subclass
            raise NotImplementedError
    """
    def __call__(self, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        assert hasattr(obj, '_initialize'), \
            'AutoInitializeMeta requires that subclass implements _initialize()'
        obj._initialize()
        return obj


class noop_context:
    """
    Placeholder context manager that does nothing.
    We could have written simply as:

    @contextmanager
    def noop_context(*args, **kwargs):
        yield

    but the returned context manager cannot be called twice, i.e.
    my_noop = noop_context()
    with my_noop:
        do1()
    with my_noop: # trigger generator error
        do2()
    """
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def meta_wrap(decor):
    """
    a decorator decorator, allowing the wrapped decorator to be used as:
    @decorator(*args, **kwargs)
    def callable()
      -- or --
    @decorator  # without parenthesis, args and kwargs will use default
    def callable()

    Args:
      decor: a decorator whose first argument is a callable (function or class
        to be decorated), and the rest of the arguments can be omitted as default.
        decor(f, ... the other arguments must have default values)

    Warning:
      decor can NOT be a function that receives a single, callable argument.
      See stackoverflow: http://goo.gl/UEYbDB
    """
    single_callable = (lambda args, kwargs:
                       len(args) == 1 and len(kwargs) == 0 and callable(args[0]))
    @functools.wraps(decor)
    def new_decor(*args, **kwargs):
        if single_callable(args, kwargs):
            # this is the double-decorated f.
            # It should not run on a single callable.
            return decor(args[0])
        else:
            # decorator arguments
            return lambda real_f: decor(real_f, *args, **kwargs)

    return new_decor


@meta_wrap
def deprecated(func, msg='', action='warning'):
    """
    Function/class decorator: designate deprecation.

    Args:
      msg: string message.
      action: string mode
      - 'warning': (default) prints `msg` to stderr
      - 'noop': do nothing
      - 'raise': raise DeprecatedError(`msg`)
    """
    action = action.lower()
    if action not in ['warning', 'noop', 'raise']:
        raise ValueError('unknown action type {}'.format(action))
    if not msg:
        msg = 'This is a deprecated feature.'

    # only does the deprecation when being called
    @functools.wraps(func)
    def _deprecated(*args, **kwargs):
        if action == 'warning':
            print(msg, file=sys.stderr)
        elif action == 'raise':
            raise DeprecationWarning(msg)
        return func(*args, **kwargs)
    return _deprecated


class ArgParser(object):
    def __init__(self, **kwargs):
        """
        The following options are pre-configured
        --verbosity, or -vvv (number of v's indicate the level of verbosity)
        --debug: turn on debugging mode
        """
        kwargs['formatter_class'] = ArgParser._SingleMetavarFormatter
        self.parser = argparse.ArgumentParser(**kwargs)
        self.parser.add_argument('--verbose', '-v', action='count', default=-1,
                                 help='can repeat, e.g. -vvv for level 3 verbosity')
        self.parser.add_argument('--debug', action='store_true', default=False,
                                 help='Turn on debugging mode. ')

    def add(self, *args, **kwargs):
        default = kwargs.get('default')
        dtype = kwargs.get('type')
        if dtype is None:
            if default is None:
                dtype = str
            else:
                dtype = type(default)
        typename = dtype.__name__
        if 'metavar' not in kwargs:
            # metavar: display --foo <float=0.05> in help string
            if 'choices' in kwargs:
                choices = kwargs['choices']
                choices_str = '/'.join(['{}']*len(choices)).format(*choices)
                kwargs['metavar'] = '<{}: {}>'.format(typename, choices_str)
            elif 'nargs' in kwargs:
                # better formatting handled in _SingleMetavarFormatter
                kwargs['metavar'] = '{}'.format(typename)
            elif not kwargs.get('action'):
                # if 'store_true', then no metavar needed
                # list of actions: https://docs.python.org/3/library/argparse.html#action
                default_str = '={}'.format(default) if default else ''
                kwargs['metavar'] = '<{}{}>'.format(typename, default_str)
        self.parser.add_argument(*args, **kwargs)


    def add_boolean_flag(self, name, default=False, pair=True, help=None):
        """Add a boolean flag to argparse parser.

        Args:
            parser: argparse.Parser
                parser to add the flag to
            name: str
                --<name> will enable the flag, while --no-<name> will disable it
            default: bool or None
                default value of the flag
            pair:
                True to add both --myflag and --no-myflag
            help: str
                help string for the flag
        """
        self.parser.add_argument("--" + name,
                                 action="store_true", default=default, help=help)
        if pair:
            self.parser.add_argument("--no-" + name,
                                     action="store_false", dest=name)

    # aliases
    add_argument = add
    def parse(self, *args, **kwargs):
        return self.parser.parse_args(*args, **kwargs)

    def __getattr__(self, attr):
        "delegate any other methods to the underlying parser"
        if attr in dir(self):
            return object.__getattribute__(self, attr)
        else:
            return getattr(self.parser, attr)

    class _SingleMetavarFormatter(argparse.HelpFormatter):
        "Helper for better metavar display in ArgParser"
        def _format_action_invocation(self, action):
            if not action.option_strings:
                metavar, = self._metavar_formatter(action, action.dest)(1)
                return metavar
            else:
                parts = []
                # if the Optional doesn't take a value, format is `-s, --long`
                if action.nargs == 0:
                    parts.extend(action.option_strings)
                # if the Optional takes a value, format is:
                #    -s <METAVAR>, --long
                else:
                    default = action.dest.upper()
                    args_string = self._format_args(action, default)
                    ## THIS IS THE PART REPLACED
                    # for option_string in action.option_strings:
                    # parts.append('%s %s' % (option_string, args_string))
                    parts.extend(action.option_strings)
                    # treat nargs different
                    if action.nargs and action.default:
                        parts[-1] += ' default={}'.format(action.default)
                    parts[0] += ' ' + args_string
                return ', '.join(parts)

class MovingAverageRecorder():
    """
        Records moving average 
    """
    def __init__(self, decay=0.95):
        self.decay = decay
        self.cum_value = 0
        self.normalization = 0

    def add_value(self, value):
        self.cum_value *= self.decay
        self.cum_value += value

        self.normalization *= self.decay
        self.normalization += 1

        return self.cum_value / self.normalization

    def cur_value(self):
        """
            Returns current moving average, 0 if no records
        """
        if self.normalization == 0:
            return 0
        else:
            return self.cum_value / self.normalization

class ThreadSafeMovingAverageRecorder(MovingAverageRecorder):
    def __init__(self, decay=0.95):
        super().__init__()
        self.lock = Lock()

    def add_value(self, value):
        with self.lock:
            return super().add_value(value)

    def cur_value(self):
        with self.lock:
            return super().cur_value()

class TimeRecorder():
    """
        Records average of whatever context block it is recording
        Don't call time in two threads
    """
    def __init__(self, decay=0.9995, max_seconds=10):
        """
        Args:
            decay: Decay factor of smoothed moving average
                    Default is 0.9995, which is approximately moving average 
                    of 2000 samples
            max_seconds: round down all time differences larger than specified
                        Useful when the application just started and there are long waits 
                        that might throw off the average
        """
        self.moving_average = ThreadSafeMovingAverageRecorder(decay)
        self.max_seconds = max_seconds
        self.started = False

    @contextmanager
    def time(self):
        pre_time = time.time()
        yield None
        post_time = time.time()

        interval = min(self.max_seconds, post_time - pre_time)
        self.moving_average.add_value(interval)

    def start(self):
        if self.started:
            raise RuntimeError('Starting a started timer')
        self.pre_time = time.time()
        self.started = True

    def stop(self):
        if not self.started:
            raise RuntimeError('Stopping a timer that is not started')
        self.post_time = time.time()
        self.started = False
    
        interval = min(self.max_seconds, self.post_time - self.pre_time)
        self.moving_average.add_value(interval)

    def lap(self):
        if not self.started:
            raise RuntimeError('Stopping a timer that is not started')
        post_time = time.time()
    
        interval = min(self.max_seconds, post_time - self.pre_time)
        self.moving_average.add_value(interval)

        self.pre_time = post_time


    @property
    def avg(self):
        return self.moving_average.cur_value()


class PeriodicWakeUpWorker(Thread):
    """
    Args:
        @target: The function to be called periodically
        @interval: Time between two calls 
        @args: Args to feed to target()
        @kwargs: Key word Args to feed to target()
    """
    def __init__(self, target, interval=1, args=None, kwargs=None):
        Thread.__init__(self)
        self.target = target
        self.interval = interval
        self.args = args
        self.kwargs = kwargs

    def run(self):
        if self.args is None:
            self.args = []
        if self.kwargs is None:
            self.kwargs = {}
        while True:
            self.target(*self.args, **self.kwargs)
            time.sleep(self.interval)


class TimedTracker(object):
    def __init__(self, interval):
        self.init_time = time.time()
        self.last_time = self.init_time
        self.interval = interval

    def track_increment(self):
        cur_time = time.time()
        time_since_last = cur_time - self.last_time
        enough_time_passed = time_since_last >= self.interval
        if enough_time_passed:
            self.last_time = cur_time
        return enough_time_passed


class AverageValue(object):
    """
        Keeps track of average of things
        Always caches the latest value so no division by 0
    """
    def __init__(self, initial_value):
        self.last_val = initial_value
        self.sum = initial_value
        self.count = 1

    def add(self, value):
        self.last_val = value
        self.sum += value
        self.count += 1

    def avg(self, clear=True):
        """
            Get the average of the currently tracked value
        Args:
            @clear: if true (default), clears the cached sum/count
        """
        ans = self.sum / self.count
        if clear:
            self.sum = self.last_val
            self.count = 1
        return ans


class AverageDictionary(object):
    def __init__(self):
        self.data = {}

    def add_scalars(self, new_data):
        for key in new_data:
            if key in self.data:
                self.data[key].add(new_data[key])
            else:
                self.data[key] = AverageValue(new_data[key])

    def get_values(self, clear=True):
        response = {}
        for key in self.data:
            response[key] = self.data[key].avg(clear=clear)
        return response


def wait_for_popen(processes, verbose=True):
    """

    Wait for some processes, terminate all if one of them fails

    Args:
        processes: A list of Popen
        verbose (bool, default True): when True, print to console
            when a process exited with code 0

    Raises:
        RuntimeError: When one process fails
    """
    completed = [False for i in processes]
    while True:
        time.sleep(1)
        for i, process in enumerate(processes):
            if completed[i]:
                continue
            ret = process.poll()
            if ret is not None:
                if ret == 0:
                    completed[i] = True
                    if verbose:
                        print('Process {} exited with code 0'.format(i))
                else:
                    for process in processes:
                        if process is not None:
                            process.kill()
                    raise RuntimeError('Process {} exited with code {}'
                                       .format(i, ret))


def start_thread(func, daemon=True, args=None, kwargs=None):
    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}
    t = Thread(
        target=func,
        args=args,
        kwargs=kwargs,
        daemon=daemon,
    )
    t.start()
    return t



import os
import sys
import psutil
import subprocess


# Currently planned tests
# DDPG dm_control
# DDPG mujocomanip
# PPO dm_control
# PPO mujocomanip


def _setup_env():
    """
    Setup the necessary environment variables
    """
    os.environ["SYMPH_PS_BACKEND_PORT"] = "7006"
    os.environ["SYMPH_PARAMETER_PUBLISH_PORT"] = "7001"
    os.environ["SYMPH_SAMPLER_FRONTEND_ADDR"] = "7004"
    os.environ["SYMPHONY_PARAMETER_SERVER_HOST"] = "127.0.0.1"
    os.environ["SYMPH_TENSORPLEX_HOST"] = "127.0.0.1"
    os.environ["SYMPH_TENSORPLEX_PORT"] = "7009"
    os.environ["SYMPH_LOGGERPLEX_HOST"] = "127.0.0.1"
    os.environ["SYMPH_LOGGERPLEX_PORT"] = "7003"
    os.environ["SYMPH_COLLECTOR_FRONTEND_HOST"] = "127.0.0.1"
    os.environ["SYMPH_COLLECTOR_FRONTEND_PORT"] = "7005"
    os.environ["SYMPH_PS_FRONTEND_HOST"] = "127.0.0.1"
    os.environ["SYMPH_PS_FRONTEND_PORT"] = "7008"
    os.environ["SYMPH_SAMPLER_FRONTEND_HOST"] = "127.0.0.1"
    os.environ["SYMPH_SAMPLER_FRONTEND_PORT"] = "7003"
    os.environ["SYMPH_SAMPLER_BACKEND_HOST"] = "127.0.0.1"
    os.environ["SYMPH_SAMPLER_BACKEND_PORT"] = "7002"
    os.environ["SYMPH_PARAMETER_PUBLISH_HOST"] = "127.0.0.1"
    os.environ["SYMPH_PARAMETER_PUBLISH_PORT"] = "7001"
    os.environ["SYMPH_COLLECTOR_BACKEND_HOST"] = "127.0.0.1"
    os.environ["SYMPH_COLLECTOR_BACKEND_PORT"] = "7007"
    os.environ["SYMPH_PREFETCH_QUEUE_HOST"] = "127.0.0.1"
    os.environ["SYMPH_PREFETCH_QUEUE_PORT"] = "7000"


def integration_test(temp_path,
                     config_path,
                     launcher,
                     env='gym:HalfCheetah-v2',
                     additional_args=None):
    if additional_args is None:
        additional_args = []
    print("Making temp directory...")
    os.makedirs(temp_path, exist_ok=True)
    print("Setting up experiment launcher...")
    args = [
        '--unit-test',
        '--num-agents',
        '1',
        '--env',
        'gym:HalfCheetah-v2',
        # 'robosuite:SawyerLift',
        # 'dm_control:cartpole-balance',
        '--experiment-folder',
        str(temp_path)] + additional_args

    print("Setting up environment variables...")
    _setup_env()

    subprocesses = []

    for module in ['replay', 'ps', 'eval-0']:  # tensorboard,
        subprocesses.append(subprocess.Popen([sys.executable,
                                              '-u',
                                              config_path,
                                              module,
                                              '--'] + args))
        print(module + '=' * 20 + 'done')
    print('Supplementary components launched')

    launcher.setup(args)

    print('Launcher setup')

    agent = launcher.setup_agent(0)
    agent.main_setup()

    print('Agent setup')

    learner = launcher.setup_learner()
    learner.main_setup()

    print('Learner setup')

    for i in range(2):
        print('Iteration {}'.format(i))
        for j in range(2):
            agent.main_loop()
        learner.main_loop()

    for subprocess_ in subprocesses:
        print(psutil.Process(subprocess_.pid).status())
        # assert(psutil.Process(subprocess_.pid).status() == 'running')

    parent = psutil.Process()
    for child in parent.children(recursive=True):
        child.terminate()

    print('Finished testing.')



import os
import queue
from caraml.zmq import DataFetcher
from benedict import BeneDict
import surreal.utils as U
from threading import Thread


class LearnerDataPrefetcher(DataFetcher):
    """
        Convenience class that initializes everything from session config
        + batch_size

        Fetches data from replay in multiple processes and put them into
        a queue
    """
    def __init__(self,
                 session_config,
                 batch_size,
                 worker_preprocess=None,
                 main_preprocess=None):
        self.max_fetch_queue = session_config.learner.max_prefetch_queue
        self.max_preprocess_queue = session_config.learner.max_preprocess_queue
        self.fetch_queue = queue.Queue(maxsize=self.max_fetch_queue)
        self.preprocess_queue = queue.Queue(maxsize=self.max_preprocess_queue)
        self.timer = U.TimeRecorder()

        self.sampler_host = os.environ['SYMPH_SAMPLER_FRONTEND_HOST']
        self.sampler_port = os.environ['SYMPH_SAMPLER_FRONTEND_PORT']
        self.batch_size = batch_size
        self.prefetch_processes = session_config.learner.prefetch_processes
        self.prefetch_host = '127.0.0.1'
        self.worker_comm_port = os.environ['SYMPH_PREFETCH_QUEUE_PORT']
        self.worker_preprocess = worker_preprocess
        self.main_preprocess = main_preprocess
        super().__init__(
            handler=self._put,
            remote_host=self.sampler_host,
            remote_port=self.sampler_port,
            requests=self.request_generator(),
            worker_comm_port=self.worker_comm_port,
            remote_serializer=U.serialize,
            remote_deserialzer=U.deserialize,
            n_workers=self.prefetch_processes,
            worker_handler=self.worker_preprocess)

    def run(self):
        self._preprocess_thread = Thread(target=self._preprocess_loop,
                                         daemon=True)
        self._preprocess_thread.start()
        super().run()

    def _preprocess_loop(self):
        while True:
            sharedmem_obj = self.fetch_queue.get(block=True)
            batch = BeneDict(sharedmem_obj.data)
            batch = self.main_preprocess(batch)
            self.preprocess_queue.put(batch)

    def _put(self, _, data):
        self.fetch_queue.put(data, block=True)

    def get(self):
        """
            Returns a SharedMemoryObject
            whose .data attribute contains data
        """
        with self.timer.time():
            return self.preprocess_queue.get(block=True)

    def request_generator(self):
        while True:
            yield self.batch_size



import numpy as np
from collections import deque
import gym
from gym import spaces
# import cv2
from .atari_names import atari_name_cap


"""
If `GLIBCXX_3.4.20` not found for Atari, run
`conda install libgcc`
https://github.com/openai/gym/issues/543
"""
class NoopResetEnv(gym.Wrapper):
    def __init__(self, env, noop_max=30):
        """Sample initial states by taking random number of no-ops on reset.
        No-op is assumed to be action 0.
        """
        gym.Wrapper.__init__(self, env)
        self.noop_max = noop_max
        self.override_num_noops = None
        if isinstance(env.action_space, gym.spaces.MultiBinary):
            self.noop_action = np.zeros(self.env.action_space.n, dtype=np.int64)
        else:
            # used for atari environments
            self.noop_action = 0
            assert env.unwrapped.get_action_meanings()[0] == 'NOOP'

    def _reset(self, **kwargs):
        """ Do no-op action for a number of steps in [1, noop_max]."""
        self.env.reset(**kwargs)
        if self.override_num_noops is not None:
            noops = self.override_num_noops
        else:
            noops = self.unwrapped.np_random.randint(1, self.noop_max + 1) #pylint: disable=E1101
        assert noops > 0
        obs = None
        for _ in range(noops):
            obs, _, done, _ = self.env.step(self.noop_action)
            if done:
                obs = self.env.reset(**kwargs)
        return obs

class FireResetEnv(gym.Wrapper):
    def __init__(self, env):
        """Take action on reset for environments that are fixed until firing."""
        gym.Wrapper.__init__(self, env)
        assert env.unwrapped.get_action_meanings()[1] == 'FIRE'
        assert len(env.unwrapped.get_action_meanings()) >= 3

    def _reset(self, **kwargs):
        self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(1)
        if done:
            self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(2)
        if done:
            self.env.reset(**kwargs)
        return obs

class EpisodicLifeEnv(gym.Wrapper):
    def __init__(self, env):
        """Make end-of-life == end-of-episode, but only reset on true game over.
        Done by DeepMind for the DQN and co. since it helps value estimation.
        """
        gym.Wrapper.__init__(self, env)
        self.lives = 0
        self.was_real_done  = True

    def _step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.was_real_done = done
        # check current lives, make loss of life terminal,
        # then update lives to handle bonus lives
        lives = self.env.unwrapped.ale.lives()
        if lives < self.lives and lives > 0:
            # for Qbert somtimes we stay in lives == 0 condtion for a few frames
            # so its important to keep lives > 0, so that we only reset once
            # the environment advertises done.
            done = True
        self.lives = lives
        return obs, reward, done, info

    def _reset(self, **kwargs):
        """Reset only when lives are exhausted.
        This way all states are still reachable even though lives are episodic,
        and the learner need not know about any of this behind-the-scenes.
        """
        if self.was_real_done:
            obs = self.env.reset(**kwargs)
        else:
            # no-op step to advance from terminal/lost life state
            obs, _, _, _ = self.env.step(0)
        self.lives = self.env.unwrapped.ale.lives()
        return obs

class MaxAndSkipEnv(gym.Wrapper):
    def __init__(self, env, skip=4):
        """Return only every `skip`-th frame"""
        gym.Wrapper.__init__(self, env)
        # most recent raw observations (for max pooling across time steps)
        self._obs_buffer = np.zeros((2,)+env.observation_space.shape, dtype='uint8')
        self._skip       = skip

    def _step(self, action):
        """Repeat action, sum reward, and max over last observations."""
        total_reward = 0.0
        done = None
        for i in range(self._skip):
            obs, reward, done, info = self.env.step(action)
            if i == self._skip - 2: self._obs_buffer[0] = obs
            if i == self._skip - 1: self._obs_buffer[1] = obs
            total_reward += reward
            if done:
                break
        # Note that the observation on the done=True frame
        # doesn't matter
        max_frame = self._obs_buffer.max(axis=0)

        return max_frame, total_reward, done, info

class ClipRewardEnv(gym.RewardWrapper):
    def _reward(self, reward):
        """Bin reward to {+1, 0, -1} by its sign."""
        return np.sign(reward)

# class WarpFrame(gym.ObservationWrapper):
#     def __init__(self, env, is_pytorch=True):
#         """Warp frames to 84x84 as done in the Nature paper and later work."""
#         gym.ObservationWrapper.__init__(self, env)
#         self.width = 84
#         self.height = 84
#         shape = (self.width, self.height)
#         shape = (1,) + shape if is_pytorch else shape + (1,)
#         self.observation_space = spaces.Box(low=0, high=255, shape=shape )
#         self.is_pytorch = is_pytorch

#     def _observation(self, frame):
#         frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
#         frame = cv2.resize(
#             frame,
#             (self.width, self.height),
#             interpolation=cv2.INTER_AREA
#         )
#         return frame[:, :, None] if self.is_pytorch else frame[None, :, :]


class FrameStack(gym.Wrapper):
    def __init__(self, env, k, lazy, is_pytorch=True):
        """Stack k last frames.

        Returns lazy array, which is much more memory efficient.

        See Also
        --------
        baselines.common.atari_wrappers.LazyFrames
        """
        gym.Wrapper.__init__(self, env)
        self.k = k
        self.frames = deque([], maxlen=k)
        shp = env.observation_space.shape
        self.observation_space = spaces.Box(low=0, high=255, shape=(shp[0], shp[1], shp[2] * k))
        self.lazy = lazy
        self.is_pytorch = is_pytorch

    def _reset(self):
        ob = self.env.reset()
        for _ in range(self.k):
            self.frames.append(ob)
        info = {}
        info['frames'] = list(self.frames)
        return self._get_ob(), info

    def _step(self, action):
        ob, reward, done, info = self.env.step(action)
        self.frames.append(ob)
        # return individual observations in info dict
        # don't forget to delete them before sending to replay!
        info['frames'] = list(self.frames)
        return self._get_ob(), reward, done, info

    def _get_ob(self):
        assert len(self.frames) == self.k
        obs = LazyFrames(list(self.frames), is_pytorch=self.is_pytorch)
        if self.lazy:
            return obs
        else:
            return obs.numpy()


class ScaledFloatFrame(gym.ObservationWrapper):
    def _observation(self, observation):
        # careful! This undoes the memory optimization, use
        # with smaller replay buffers only.
        return np.array(observation).astype(np.float32) / 255.0


class LazyFrames(object):
    def __init__(self, frames, is_pytorch=True):
        """This object ensures that common frames between the observations are only stored once.
        It exists purely to optimize memory usage which can be huge for DQN's 1M frames replay
        buffers.

        This object should only be converted to numpy array before being passed to the model.

        You'd not belive how complex the previous solution was."""
        self._frames = frames
        self.is_pytorch = is_pytorch

    def numpy(self, dtype=None):
        obs = np.concatenate(
            self._frames,
            axis=0 if self.is_pytorch else 2
        )
        if dtype is not None:
            obs = obs.astype(dtype)
        return obs

    def __array__(self, dtype=None):
        return self.numpy(dtype)


def make_atari(name):
    env = gym.make(atari_name_cap(name) + "NoFrameskip-v4")
    return env


def wrap_deepmind(env,
                  episode_life=True,
                  clip_rewards=True,
                  lazy_frame=False,
                  scale=False):
    """Configure environment for DeepMind-style Atari.
    """
    env = NoopResetEnv(env, noop_max=30)
    env = MaxAndSkipEnv(env, skip=4)
    if episode_life:
        env = EpisodicLifeEnv(env)
    if 'FIRE' in env.unwrapped.get_action_meanings():
        env = FireResetEnv(env)
    env = WarpFrame(env)
    if scale:
        env = ScaledFloatFrame(env)
    if clip_rewards:
        env = ClipRewardEnv(env)
    env = FrameStack(env, 4, lazy=lazy_frame)
    return env




import os
import time

from .wrapper import Wrapper
import imageio

from multiprocessing import Process, Queue
import numpy as np

# File taken from 5d4757b9d2789c34a1120f4c6e4452e4cb7a11fc by Zihua

def save_video(frame_queue, filename, fps):
    '''
    Target function for video process. Opens up file with path and uses library imageio
    Args:
        frame_queue: a queue of frames to be store. If the frame is None, the capture is over
        filename: filename to which the capture is to be stored
        fps: framerate.
    Note that Queue.get() is a blocking function and thus will hang until new frames are
    added to frame_queue
    '''
    writer = imageio.get_writer(filename, fps=fps)
    while True:
        frame = frame_queue.get()
        if frame is None:
            break
        writer.append_data(frame)
    writer.close()

class VideoWrapper(Wrapper):
    '''
    Environment Wrappers for automatically rendering and saving test runs.
    Attributes:
        public attributes
            env (Env): environment to be wrapped
            capture_interval (int): number of episodes between captures
                default to 10
            frame_interval (int): number of frames between each recorded frame
                default to 10
            fps (int): frame rate
                default to 30
            ext (str): file extention. Either .gif or .mp4 depending on input
                default to .mp4
            save_folder (str): directory to which the files are to be saved
                default to 'main-experiment-folder/videos'
            max_videos (int): maximum number of videos allowed in a directory
                default to 10
        helper attributes
            num_eps (int): number of episodes executed
            num_steps (int): number of steps executed
            is_recording (bool): whether the on going episode is being recorded
            path_queue (Queue): to keep track of saved videos to maintain max number of captures
            num_paths (int): number of paths stored in path_queue. necessary because Queue.qsize()
                             is not implemented on Mac.
            video_process (Process): separate process that writes images to file
            video_queue (Queue): queue of frames to be writen to file
    '''

    def __init__(self, env, env_config, session_config,
        frame_interval = 10, fps =30, use_gif = False):
        '''
        Constructor for VideoWrapper. also creates the save directory if not present
        Args:
            env (Env): environment to be wrapped
            capture_interval (int): number of episodes between captures
            frame_interval (int): number of frames between each recorded frame
            fps (int): frame rate
            save_folder (str): directory to which the files are to be saved
            max_videos (int): maximum number of videos allowed in a directory
            use_gif (bool): boolean flag to use either gif or mp4
        '''

        super().__init__(env)
        self.env = env
        self.env_category = env_config.env_name.split(':')[0]

        self.max_videos = env_config.video.max_videos
        self.capture_interval = env_config.video.record_every
        self.frame_interval   = frame_interval
        self.fps = fps

        self.ext = '.gif' if use_gif else '.mp4'
        self.save_folder = env_config.video.save_folder
        if not self.save_folder:
            self.save_folder = os.path.join(session_config.folder, 'videos')
        self.save_folder = os.path.expanduser(self.save_folder)

        if not os.path.exists(self.save_folder):
            os.makedirs(self.save_folder)

        self.num_eps = 0
        self.num_steps = 0
        self.is_recording = False
        self.path_queue = Queue()
        self.num_paths = 0 #work around, qsize() not implemented on mac

    def _reset(self, **kwargs):
        '''
        Overwrites reset method. in addition to reseting the environment,
        this method also manages the video process and when to start writing video
        or gif to file
        '''
        self.num_steps = 0

        if self.is_recording:
            self.stop_record()

        if self.num_eps % self.capture_interval == 0:
            self.video_queue = Queue()
            self.is_recording = True

            path = os.path.join(self.save_folder, 'video_eps_{}{}'.format(self.num_eps, self.ext))
            if self.num_paths >= self.max_videos:
                dep_path = self.path_queue.get()
                os.remove(dep_path)
                self.num_paths -= 1

            self.path_queue.put(path)
            self.video_process = Process(target=save_video,
                                         args=(self.video_queue, path, self.fps))
            self.video_process.start()
            self.num_paths += 1

        state = self.env.reset(**kwargs)
        self.num_eps += 1
        return state

    def _step(self, action):
        '''
        Overwrites _step function. In addition to taking an action,
        if the video is recording and its time to capture a frame, the
        frame is rendered using 'rgb_array' mode and is put into the video_queue.
        If video capture is over, stop_recording is called.
        '''
        state, step_reward, terminal, info = self.env.step(action)
        self.num_steps += 1
        if self.is_recording:
            # For dm_control videos, the video will be transposed when we save to disk
            # We correct for that here
            ob = self.env.render()
            if self.env_category == 'dm_control':
                ob = ob.transpose(1, 0, 2)
            else: # mujocomanip
                ob = np.rot90(ob, 2)
            self.video_queue.put(ob)

        return state, step_reward, terminal, info

    def stop_record(self):
        '''
        stops recording and wait for the video_process to finish writing to file and join
        First puts a None (End of Video) frame into the frame queue and wait for writing
        process to terminate.
        '''
        self.video_queue.put(None)
        self.video_process.join()
        self.is_recording = False
        print('finished recording video {}'.format(self.num_eps))




from .common import *
from .numpy_util import *
from .schedule import *
from .filesys import *
from .serializer import *
from .checkpoint import Checkpoint, PeriodicCheckpoint
from .config_yaml import get_config_file


import torch.nn as nn
from torch.nn.init import xavier_uniform
import surreal.utils as U
import torch.nn.functional as F
import numpy as np
import torchx.nn as nnx


class DuelingQbase(nnx.Module):
    def init_dueling(self, *,
                     action_dim,
                     prelinear_size,
                     fc_hidden_sizes,
                     dueling):
        """
        Args:
            - prelinear_size: size of feature vector before the linear layers,
                like flattened conv or LSTM features
            - fc_hidden_sizes: list of fully connected layer sizes before `action_dim` softmax
        """
        self.dueling = dueling
        self.prelinear_size = prelinear_size
        U.assert_type(fc_hidden_sizes, list)
        hiddens = [prelinear_size] + fc_hidden_sizes
        self.fc_action_layers = nn.ModuleList()
        hidden_list = hiddens + [action_dim]
        for prev_h, next_h in zip(hidden_list[:-1], hidden_list[1:]):
            lin = nn.Linear(prev_h, next_h)
            U.conv_fc_init(lin)
            self.fc_action_layers.append(lin)

        if dueling:
            self.fc_state_layers = nn.ModuleList()
            # output a single state value
            hidden_list = hiddens + [1]
            for prev_h, next_h in zip(hidden_list[:-1], hidden_list[1:]):
                lin = nn.Linear(prev_h, next_h)
                U.conv_fc_init(lin)
                self.fc_state_layers.append(lin)

    def forward(self, x):
        """
        x is the processed tensor from raw states before the dueling layers,
        its shape should be [batch x prelinear_size]
        """
        shape = x.size()
        assert len(shape) == 2 and shape[1] == self.prelinear_size

        action_scores = x
        for is_last, fc_action in U.iter_last(self.fc_action_layers):
            action_scores = fc_action(action_scores)
            if not is_last:
                action_scores = F.relu(action_scores)

        if self.dueling:
            action_scores_mean = action_scores.mean(dim=1, keepdim=True)
            action_scores -= action_scores_mean
            state_score = x
            for is_last, fc_action in U.iter_last(self.fc_state_layers):
                state_score = fc_action(state_score)
                if not is_last:
                    state_score = F.relu(state_score)
            return action_scores + state_score
        else:
            return action_scores


class FFQfunc(DuelingQbase):
    def __init__(self, *,
                 input_shape,
                 action_dim,
                 convs,
                 fc_hidden_sizes,
                 dueling,
                 is_uint8=False):
        """
        is_uint8: scale down uint8 image input from Atari by 255.0
        You can use atari_wrapper.ScaledFloatFrame, but dividing on GPU is more
        computationally and memory efficient
        """
        super().__init__()
        if convs:
            assert len(input_shape) == 3
        self.is_uint8 = is_uint8
        shape = input_shape
        in_channel = input_shape[0]
        self.conv_layers = nn.ModuleList()
        for out_channel, kernel_size, stride in convs:
            # MUST pad to make 'SAME' mode (TF default), otherwise learning much worse
            # pytorch defaults to 'VALID' mode, i.e. pad=0
            conv = nn.Conv2d(in_channel,
                             out_channel,
                             kernel_size,
                             stride=stride,
                             padding=kernel_size // 2)
            U.conv_fc_init(conv)
            in_channel = out_channel
            self.conv_layers.append(conv)
            shape = U.infer_shape_conv2d(shape, out_channel,
                                         kernel_size,
                                         stride=stride,
                                         padding=kernel_size // 2)

        if not U.is_valid_shape(shape):
            raise ValueError('Pre-FC shape has invalid size: {}'.format(shape))

        self.init_dueling(action_dim=action_dim,
                          prelinear_size=int(np.prod(shape)),
                          fc_hidden_sizes=fc_hidden_sizes,
                          dueling=dueling)

    def forward(self, x):
        if self.is_uint8:
            # do NOT use .div_, because the observation may be reused later
            x = x / 255.0
        for conv in self.conv_layers:
            x = F.relu(conv(x))
        x = U.flatten_conv(x)
        return super().forward(x)


def build_ffqfunc(learner_config, env_config):
    action_dim = env_config.action_spec.dim
    assert len(action_dim) == 1
    assert env_config.action_spec.type == 'discrete'
    action_dim = action_dim[0]
    q_func = FFQfunc(
        input_shape=env_config.obs_spec.dim,
        action_dim=action_dim,
        convs=learner_config.model.convs,
        fc_hidden_sizes=learner_config.model.fc_hidden_sizes,
        dueling=learner_config.model.dueling
    )
    return q_func, action_dim



import numpy as np
# Based on https://github.com/openai/baselines/blob/bb403781182c6e31d3bf5de16f42b0cb0d8421f7/baselines/ddpg/noise.py

class ActionNoise(object):
    def reset(self):
        pass


class NormalActionNoise(ActionNoise):
    def __init__(self, mu, sigma):
        self.mu = mu
        self.sigma = sigma

    def __call__(self):
        return np.random.normal(self.mu, self.sigma)

    def __repr__(self):
        return 'NormalActionNoise(mu={}, sigma={})'.format(self.mu, self.sigma)


# Based on http://math.stackexchange.com/questions/1287634/implementing-ornstein-uhlenbeck-in-matlab
class OrnsteinUhlenbeckActionNoise(ActionNoise):
    def __init__(self, mu, sigma, theta, dt, x0=None):
        self.theta = theta
        self.mu = mu
        self.sigma = sigma
        self.dt = dt
        self.x0 = x0
        self.reset()

    def __call__(self):
        x = self.x_prev + self.theta * (self.mu - self.x_prev) * self.dt + self.sigma * np.sqrt(self.dt) * np.random.normal(size=self.mu.shape)
        self.x_prev = x
        return x

    def reset(self):
        self.x_prev = self.x0 if self.x0 is not None else np.zeros_like(self.mu)

    def __repr__(self):
        return 'OrnsteinUhlenbeckActionNoise(mu={}, sigma={})'.format(self.mu, self.sigma)





import torch
import surreal.utils as U
#from surreal.utils.pytorch import GpuVariable as Variable
from surreal.session import PeriodicTracker
from surreal.model.q_net import build_ffqfunc
from .base import Learner


class DQNLearner(Learner):
    def __init__(self, learner_config, env_config, session_config):
        super().__init__(learner_config, env_config, session_config)
        self.q_func, self.action_dim = build_ffqfunc(
            self.learner_config,
            self.env_config
        )
        self.algo = self.learner_config.algo
        self.q_target = self.q_func.clone()
        self.optimizer = torch.optim.Adam(
            self.q_func.parameters(),
            lr=self.algo.lr,
            eps=1e-4
        )
        self.target_update_tracker = PeriodicTracker(
            period=self.algo.target_network_update_freq,
        )

    def _update_target(self):
        self.q_target.copy_from(self.q_func)

    def _run_optimizer(self, loss):
        self.optimizer.zero_grad()
        loss.backward()
        norm_clip = self.algo.grad_norm_clipping
        if norm_clip is not None:
            self.q_func.clip_grad_norm(norm_clip)
            # torch.nn.utils.net_clip_grad_norm(
            #     self.q_func.parameters(),
            #     max_norm=norm_clip
            # )
        self.optimizer.step()

    def _optimize(self, obs, actions, rewards, obs_next, dones, weights):
        # Compute Q(s_t, a)
        # columns of actions taken
        batch_size = obs.size(0)
        assert (U.shape(actions)
                == U.shape(rewards)
                == U.shape(dones)
                == (batch_size, 1))
        q_t_at_action = self.q_func(obs).gather(1, actions)
        q_tp1 = self.q_target(obs_next)
        # Double Q
        if self.algo.double_q:
            # select argmax action using online weights instead of q_target
            q_tp1_online = self.q_func(obs_next)
            q_tp1_online_argmax = q_tp1_online.max(1, keepdim=True)[1]
            q_tp1_best = q_tp1.gather(1, q_tp1_online_argmax)
        else:
            # Minh 2015 Nature paper
            # use target network for both policy and value selection
            q_tp1_best = q_tp1.max(1, keepdim=True)[0]
        # Q value for terminal states are 0
        q_tp1_best = (1.0 - dones) * q_tp1_best
        # .detach() stops gradient and makes the Variable forget its creator
        q_tp1_best = q_tp1_best.detach()
        # RHS of bellman equation
        q_expected = rewards + self.algo.gamma * q_tp1_best
        td_error = q_t_at_action - q_expected
        # torch_where
        raw_loss = U.huber_loss_per_element(td_error)
        weighted_loss = torch.mean(weights * raw_loss)
        self._run_optimizer(weighted_loss)
        return td_error

    def learn(self, batch_exp):
        weights = (U.torch_ones_like(batch_exp.rewards))
        td_errors = self._optimize(
            batch_exp.obs,
            batch_exp.actions,
            batch_exp.rewards,
            batch_exp.obs_next,
            batch_exp.dones,
            weights,
        )
        batch_size = batch_exp.obs.size(0)
        if self.target_update_tracker.track_increment(batch_size):
            # Update target network periodically.
            self._update_target()
        mean_td_error = U.to_scalar(torch.mean(torch.abs(td_errors)))
        self.tensorplex.add_scalars({
            'td_error': mean_td_error
        })

    def default_config(self):
        return {
            'model': {
                'convs': '_list_',
                'fc_hidden_sizes': '_list_',
                'dueling': '_bool_'
            },
            'algo': {
                'lr': 1e-3,
                'optimizer': 'Adam',
                'grad_norm_clipping': 10,
                'gamma': .99,
                'target_network_update_freq': '_int_',
                'double_q': True,
                'exploration': {
                    'schedule': 'linear',
                    'steps': '_int_',
                    'final_eps': 0.01,
                },
                'prioritized': {
                    'enabled': False,
                    'alpha': 0.6,
                    'beta0': 0.4,
                    'beta_anneal_iters': None,
                    'eps': 1e-6
                },
            },
        }

    def module_dict(self):
        return {
            'q_func': self.q_func
        }

    """
    def train_batch(self, batch_i, exp):
        C = self.config

        if C.prioritized.enabled:
            # TODO
            replay_buffer = PrioritizedReplayBuffer(C.buffer_size,
                                                alpha=C.prioritized.alpha,
                                                obs_encode_mode=C.obs_encode_mode)
            beta_iters = C.prioritized.beta_anneal_iters
            if beta_iters is None:
                beta_iters = C.max_timesteps
            beta_schedule = U.LinearSchedule(schedule_timesteps=beta_iters,
                                             initial_p=C.prioritized.beta0,
                                             final_p=1.0)
        else:
            beta_schedule = None

        # TODO train_freq removed, is it useful at all?
        # Minimize the error in Bellman's equation on a batch sampled from replay buffer.
        if C.prioritized.enabled:
            experience = self.replay.sample(C.batch_size,
                                              beta=beta_schedule.value(T))
            (obs, actions, rewards, obs_next, dones, weights, batch_idxes) = experience
        else:
            weights = Variable(U.torch_ones_like(exp.rewards))
            batch_idxes = None

        td_errors = self.optimize(
            exp.obs[0],
            exp.actions,
            exp.rewards,
            exp.obs[1],
            exp.dones,
            weights,
        )
        batch_size = exp.obs[0].size(0)
        if C.prioritized.enabled:
            # TODO
            new_priorities = torch.abs(td_errors) + C.prioritized.eps
            new_priorities = U.to_numpy(new_priorities)
            replay_buffer.update_priorities(batch_idxes, new_priorities)

        if self.target_update_tracker.track_increment(batch_size):
            # Update target network periodically.
            self._update_target()

    """



"""
    Utilities for building custom docker images
"""
from symphony.addons import DockerBuilder


class SurrealDockerBuilder:
    """
        Manages the following:
        1) Figure out what docker images need to be built
        2) Build (and push) them
    """

    def __init__(self,
                 build_settings,
                 images_requested,
                 tag,
                 push=False):
        """
            Figures out what docker images need to be built
            Populates self.images_provided which computes
                {name: identifier} for all images

        Args:
            build_settings (dict):
                {build_settings_name: {<docker build setting accepted by
                    symphony.addons.DockerBuilder.from_dict>}}
            images_requested (dict):
                {
                    image_name:
                    {
                        repo: image_identifier
                        build_image: build_settings_name
                    },
                    ...
                }
                Images that need to be built.

                If build_image is None, the provided image_identifier
                should be a valid docker image, DockerBuilder
                sets `image_name: image_identifier` in self.images_provided

                If build_image is not None, the provided image_identifier
                should be a valid docker image repo, DockerBuilder
                sets image_name: image_identifier+':'+tag
                in self.images_provided. It will also build the image specified
                by the build_image_field

            tag: specifies a tag for all images
            push: whether to push built images
        """
        self.build_settings = build_settings
        self.images_requested = images_requested
        self.tag = tag

        self.images_to_build = {}
        self.images_provided = {}
        for name, di in images_requested.items():
            identifier = di['identifier']
            build_config = di['build_config']
            if build_config is None:
                self.images_provided[name] = identifier
            else:
                if build_config not in self.images_to_build:
                    self.images_to_build[build_config] = []
                self.images_provided[name] = '{}:{}'.format(identifier, tag)
                self.images_to_build[build_config].append((identifier, tag))
        self.push = push

    def build(self):
        """
            Build all images and push them
        """
        for name, tag_settings in self.images_to_build.items():
            builder = DockerBuilder.from_dict(self.build_settings[name])
            builder.build()
            for repo, tag in tag_settings:
                builder.tag(repo, self.tag)
                if self.push:
                    builder.push(repo, self.tag)



import numpy as np
import torch
from surreal.utils.numpy_util import np_cast
import surreal.utils as U
import torchx.nn as nnx


class ModuleDict(object):
    """
    Two-step serialization.
    1. Each element's state_dict() is called
    2. The overall dict is then pickled.
    """
    def __init__(self, module_dict):
        U.assert_type(module_dict, dict)
        for k, m in module_dict.items():
            U.assert_type(k, str, 'Key "{}" must be string.'.format(k))
            U.assert_type(m, nnx.Module,
                          '"{}" must be torchx.nn.Module.'.format(m))
        self._module_dict = module_dict

    def dumps(self):
        """
            Dump content into binary

        Returns:
            bytes
        """
        bin_dict = {}
        for k, m in self._module_dict.items():
            state_dict = m.state_dict()
            for key in state_dict:
                state_dict[key] = state_dict[key].cpu().numpy()
            bin_dict[k] = state_dict
        return U.serialize(bin_dict)

    def loads(self, binary):
        """
            Load from binary ()

        Args:
            binary: output of ModuleDict.dumps
        """
        numpy_dict = U.deserialize(binary)
        self.load(numpy_dict)

    def load(self, numpy_dict):
        """
            Load data from a numpy_dictionary

        Args:
            numpy_dict: {
                <module_dict_key>: {
                    <paramter_key_module_dict_entries>: np.array
                }
            }
        """
        for key in numpy_dict:
            for k in numpy_dict[key]:
                numpy_dict[key][k] = torch.from_numpy(
                    np_cast(numpy_dict[key][k], np.float32))
        for k, m in self._module_dict.items():
            m.load_state_dict(numpy_dict[k])



import copy
import numpy as np
import torchx.nn as nnx

class ParameterNoise(object):
    def apply(self, model):
        pass

class NormalParameterNoise(ParameterNoise):
    def __init__(self, sigma):
        self.sigma = sigma
        print('Parameter noise initialized with sigma', self.sigma)

    def apply(self, params):
        # TODO: behavior of model.parameters() on networks with shared convolution
        for key in params:
            for k in params[key]:
                p = params[key][k]
                assert type(p) == np.ndarray
                shape = tuple(p.data.shape)
                noise = np.random.normal(0, self.sigma, size=shape)
                p = p + noise
                params[key][k] = p
        return params

    def __repr__(self):
        return 'NormalParameterNoise(sigma={})'.format(self.sigma)

class AdaptiveNormalParameterNoise(ParameterNoise):
    # Parameter noise adaptation based on https://arxiv.org/pdf/1706.01905.pdf
    def __init__(self, model_copy, module_dict_copy, target_stddev, compute_dist_interval=10, alpha=1.04, sigma=0.01):
        self.sigma = sigma
        self.target_stddev = target_stddev
        self.compute_dist_interval = compute_dist_interval
        self.alpha = alpha
        self.original_model = model_copy
        self.original_model_module_dict = module_dict_copy
        self.i = 0
        self.total_action_distance = 0.0
        print('Parameter noise initialized with sigma', self.sigma)

    def compute_action_distance(self, obs, modified_model_action):
        if self.i % self.compute_dist_interval == 0:
            # Calculate action, don't do forward pass on critic
            original_model_action, _ = self.original_model(obs, calculate_value=False)
            self.total_action_distance = (((original_model_action - modified_model_action) ** 2).sum()) ** 0.5
        self.i += 1

    def apply(self, params):
        # TODO: behavior of model.parameters() on networks with shared convolution
        if self.i > 0:
            mean_action_dist = self.total_action_distance / self.i
            print("Mean dist", mean_action_dist, "target", self.target_stddev, 'sigma', self.sigma)
            if mean_action_dist > self.target_stddev:
                self.sigma /= self.alpha
                print("Going down")
            else:
                self.sigma *= self.alpha
                print("Going up")
        self.i = 0
        # Deepcopy because module_dict converts params to tensor
        self.original_model_module_dict.load(copy.deepcopy(params))
        for key in params:
            for k in params[key]:
                p = params[key][k]
                assert type(p) == np.ndarray
                shape = tuple(p.shape)
                noise = np.random.normal(0, self.sigma, size=shape)
                p = p + noise
                params[key][k] = p
        return params

    def __repr__(self):
        return 'AdaptiveNormalParameterNoise(target={}, alpha={}, sigma={})'.format(self.target_distance, self.alpha, self.sigma)



"""
DDPG actor class
"""
import copy
import collections
import time
import torch
import numpy as np
import torchx as tx
import torchx.nn as nnx
from surreal.distributed import ModuleDict
from surreal.model.ddpg_net import DDPGModel
from surreal.env import ExpSenderWrapperSSARNStepBootstrap
from surreal.session import ConfigError
from .base import Agent
from .action_noise import *
from .param_noise import NormalParameterNoise, AdaptiveNormalParameterNoise


class DDPGAgent(Agent):
    '''
    DDPGAgent: subclass of Agent that contains DDPG algorithm logic
    Attributes:
        model: A DDPG neural network model to generate actions from observations
        noise: A traditional action noise model. When called, noise produces an output
            of the same dimension as the action dimension
        param_noise: If not None, a parameter noise model, as outlined by
            https://blog.openai.com/better-exploration-with-parameter-noise/

    Important member functions:
        public methods:
        act: method to generate action from observation using the model
        module_dict: returns the corresponding parameters
        on_parameter_fetched: performs necessary updates to parameter noise
            given new parameters from the parameter server
        pre_episode: prepares model for new episode, performs update to the noise model
    '''

    def __init__(self,
                 learner_config,
                 env_config,
                 session_config,
                 agent_id,
                 agent_mode,
                 render=False):
        '''
        Constructor for DDPGAgent class.
        Important attributes:
            learner_config, env_config, session_config: experiment configurations
            agent_id: unique id in the range [0, num_agents)
            agent_mode: toggles between agent noise and deterministic behavior
        '''
        super().__init__(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config,
            agent_id=agent_id,
            agent_mode=agent_mode,
            render=render,
        )

        self.agent_id = agent_id
        self.action_dim = self.env_config.action_spec.dim[0]
        self.obs_spec = self.env_config.obs_spec
        self.use_layernorm = self.learner_config.model.use_layernorm
        self.sleep_time = self.env_config.sleep_time

        self.param_noise = None
        self.param_noise_type = self.learner_config.algo.exploration.param_noise_type
        self.param_noise_sigma = self.learner_config.algo.exploration.param_noise_sigma
        self.param_noise_alpha = self.learner_config.algo.exploration.param_noise_alpha
        self.param_noise_target_stddev = self.learner_config.algo.exploration.param_noise_target_stddev

        self.frame_stack_concatenate_on_env = self.env_config.frame_stack_concatenate_on_env

        self.noise_type = self.learner_config.algo.exploration.noise_type

        if env_config.num_agents == 1:
            # If only one agent, we don't want a sigma of 0
            self.sigma = self.learner_config.algo.exploration.max_sigma / 3.0
        else:
            self.sigma = self.learner_config.algo.exploration.max_sigma * (
            float(agent_id) / (env_config.num_agents))
        #self.sigma = self.learner_config.algo.exploration.sigma
        print('Using exploration sigma', self.sigma)

        if torch.cuda.is_available():
            self.gpu_ids = 'cuda:all'
            self.log.info('DDPG agent is using GPU')
            # Note that user is responsible for only providing one GPU for the program
            self.log.info('cudnn version: {}'.format(torch.backends.cudnn.version()))
            torch.backends.cudnn.benchmark = True
        else:
            self.gpu_ids = 'cpu'
            self.log.info('DDPG agent is using CPU')

        with tx.device_scope(self.gpu_ids):
            self.model = DDPGModel(
                obs_spec=self.obs_spec,
                action_dim=self.action_dim,
                use_layernorm=self.use_layernorm,
                actor_fc_hidden_sizes=self.learner_config.model.actor_fc_hidden_sizes,
                critic_fc_hidden_sizes=self.learner_config.model.critic_fc_hidden_sizes,
                conv_out_channels=self.learner_config.model.conv_spec.out_channels,
                conv_kernel_sizes=self.learner_config.model.conv_spec.kernel_sizes,
                conv_strides=self.learner_config.model.conv_spec.strides,
                conv_hidden_dim=self.learner_config.model.conv_spec.hidden_output_dim,
            )
            self.model.eval()

            self._init_noise()

    def _init_noise(self):
        """
            initializes exploration noise and populates self.noise, a callable
            that returns noise of dimension same as action
        """
        if self.agent_mode == 'eval_deterministic':
            return
        if self.noise_type == 'normal':
            self.noise = NormalActionNoise(
                np.zeros(self.action_dim),
                np.ones(self.action_dim) * self.sigma
            )
        elif self.noise_type == 'ou_noise':
            self.noise = OrnsteinUhlenbeckActionNoise(
                mu=np.zeros(self.action_dim),
                sigma=self.sigma,
                theta=self.learner_config.algo.exploration.theta,
                dt=self.learner_config.algo.exploration.dt
            )
        else:
            raise ConfigError('Noise type {} undefined.'.format(self.noise_type))
        if self.param_noise_type == 'normal':
            self.param_noise = NormalParameterNoise(self.param_noise_sigma)
        elif self.param_noise_type == 'adaptive_normal':
            model_copy = copy.deepcopy(self.model)
            module_dict_copy = ModuleDict(self.module_dict(model_copy))
            self.param_noise = AdaptiveNormalParameterNoise(
                model_copy,
                module_dict_copy,
                self.param_noise_target_stddev,
                alpha=self.param_noise_alpha,
                sigma=self.param_noise_sigma
            )

    def on_parameter_fetched(self, params, info):
        params = super().on_parameter_fetched(params, info)
        if self.param_noise:
            params = self.param_noise.apply(params)
        return params

    def act(self, obs):
        with tx.device_scope(self.gpu_ids):
            if self.sleep_time > 0.0:
                time.sleep(self.sleep_time)
            if not self.frame_stack_concatenate_on_env:
                # Output pixels of environment is a list of frames,
                # we concatenate the frames into a single numpy array
                obs = copy.deepcopy(obs)
                if 'pixel' in obs:
                    for key in obs['pixel']:
                        obs['pixel'][key] = np.concatenate(obs['pixel'][key], axis=0)
            # Convert to pytorch tensor
            obs_tensor = collections.OrderedDict()
            for modality in obs:
                modality_dict = collections.OrderedDict()
                for key in obs[modality]:
                    modality_dict[key] = torch.tensor(obs[modality][key], dtype=torch.float32).unsqueeze(0)
                obs_tensor[modality] = modality_dict
            action, _ = self.model(obs_tensor, calculate_value=False)
            if self.param_noise and self.param_noise_type == 'adaptive_normal':
                self.param_noise.compute_action_distance(obs_tensor, action)
            action = action.data.cpu().numpy()[0]

            action = action.clip(-1, 1)

            if self.agent_mode != 'eval_deterministic':
                action += self.noise()

            action = action.clip(-1, 1)
            return action

    def module_dict(self, model=None):
        # By default, module_dict refers to the module_dict for the current model.
        # But, you can generate a module_dict for other models as well --
        # e.g. param_noise uses a separate module_dict to calculate action difference
        if model == None:
            model = self.model
        return {
            'ddpg': model,
        }

    def default_config(self):
        return {
            'model': {
                'convs': '_list_',
                'actor_fc_hidden_sizes': '_list_',
                'critic_fc_hidden_sizes': '_list_',
            },
        }

    def pre_episode(self):
        super().pre_episode()
        if self.agent_mode != 'eval_deterministic':
            self.noise.reset()

    def prepare_env_agent(self, env):
        env = super().prepare_env_agent(env)
        env = ExpSenderWrapperSSARNStepBootstrap(env,
                                                 self.learner_config,
                                                 self.session_config)
        return env



import itertools
from collections import deque
import surreal.utils as U
from tensorplex import TensorplexClient, LoggerplexClient
from threading import Lock
import time
import os


class PeriodicTracker(object):
    def __init__(self, period, init_value=0, init_endpoint=0):
        """
        first: if True, triggers at the first time
        """
        U.assert_type(period, int)
        assert period > 0
        U.assert_type(init_value, int)
        self.period = period
        self.value = init_value
        self._endpoint = init_endpoint

    def _update_endpoint(self):
        if self.value >= self._endpoint + self.period:
            end_incr = (self.value - self._endpoint) // self.period * self.period
            self._endpoint += end_incr
            return True
        else:
            return False

    def track_increment(self, incr=1):
        """
        Returns: True if we enter the next period
        """
        U.assert_type(incr, int)
        self.value += incr
        return self._update_endpoint()

    def track_absolute(self, value):
        """
        Returns: True if we enter the next period
        """
        U.assert_type(value, int)
        self.value = value
        return self._update_endpoint()


class RunningAverage(object):
    def __init__(self, gamma, init_value=None):
        """Keep a running estimate of a quantity. This is a bit like mean
        but more sensitive to recent changes.

        Parameters
        ----------
        gamma: float
            Must be between 0 and 1, where 0 is the most sensitive to recent
            changes.
        init_value: float or None
            Initial value of the estimate. If None, it will be set on the first update.
        """
        self._value = init_value
        self._gamma = gamma

    def update(self, new_val):
        """Update the estimate.

        Parameters
        ----------
        new_val: float
            new observated value of estimated quantity.
        """
        if self._value is None:
            self._value = new_val
        else:
            self._value = self._gamma * self._value + (1.0 - self._gamma) * new_val

    def __float__(self):
        """Get the current estimate"""
        return self._value


class TimeThrottledTensorplex(object):
    """
        A tensorplex client that aggregates output from multiple threads
        Thread safe

        Update criterion:
        1) A global step is provided when updating value
        2) Time from last update in more than min_update_interval
    """
    def __init__(self, tensorplex, min_update_interval):
        U.assert_type(tensorplex, TensorplexClient)
        self.tensorplex = tensorplex
        self.min_update_interval = min_update_interval
        self.history = U.AverageDictionary()
        self.lock = Lock()
        self.tracker = U.TimedTracker(self.min_update_interval)
        self.init_time = time.time()

    def add_scalars(self, tag_value_dict, global_step=None):
        with self.lock:
            self.history.add_scalars(tag_value_dict)
            if global_step is not None:
                if self.tracker.track_increment():
                    self.tensorplex.add_scalars(self.history.get_values(), global_step)


class PeriodicTensorplex(object):
    def __init__(self,
                 tensorplex,
                 period,
                 is_average=True,
                 keep_full_history=False):
        """
        Args:
            tensorplex: TensorplexClient object
            period: when you call `update()`, it will only send to Tensorplex
                at the specified period.
            is_average: if True, send the averaged value over the last `period`.
            keep_full_history: if False, only keep the last `period` of history.
        """
        if tensorplex is not None:  # None to turn off tensorplex
            U.assert_type(tensorplex, TensorplexClient)
        U.assert_type(period, int)
        assert period > 0
        self._tplex = tensorplex
        self._period = period
        self._is_average = is_average
        self._keep_full_history = keep_full_history
        self._tracker = PeriodicTracker(period)
        self._history = {}
        self._max_deque_size = None if keep_full_history else period

    def add_scalars(self, tag_value_dict, global_step=None):
        """

        Args:
            tag_value_dict:
            global_step: None to use the internal counter

        Returns:
            - None if period incomplete
            - dict of {tag: current_value (averaged)} at each period
        """
        for tag, value in tag_value_dict.items():
            if tag in self._history:
                self._history[tag].append(value)
            else:
                self._history[tag] = deque([value], maxlen=self._max_deque_size)
        if self._tracker.track_increment():
            current_values = {}
            for tag, history in self._history.items():
                if self._is_average:
                    history = itertools.islice(
                        history,
                        max(len(history) - self._period, 0),
                        len(history)
                    )  # simulate history[-period:] for deque
                    avg_value = U.mean(list(history))
                else:
                    avg_value = float(history[-1]) if len(history) > 0 else 0.0
                current_values[tag] = avg_value
            if self._tplex is not None:
                if global_step is None:
                    global_step = self._tracker.value
                self._tplex.add_scalars(current_values, global_step)
            return current_values

    def get_history(self):
        return {tag: list(history)
                for tag, history in self._history.items()}


def get_loggerplex_client(name, session_config):
    """
    Args:
        name: file name for the remote log file
        session_config: see session_config.loggerplex
    """
    C = session_config.loggerplex
    host = os.environ['SYMPH_LOGGERPLEX_HOST']
    port = os.environ['SYMPH_LOGGERPLEX_PORT']
    return LoggerplexClient(
        name,
        host=host,
        port=port,
        enable_local_logger=C.enable_local_logger,
        local_logger_stream='stdout',
        local_logger_level=C.local_logger_level,
        local_logger_time_format=C.local_logger_time_format,
    )


def get_tensorplex_client(client_id, session_config):
    host = os.environ['SYMPH_TENSORPLEX_HOST']
    port = os.environ['SYMPH_TENSORPLEX_PORT']
    return TensorplexClient(
        client_id,
        host=host,
        port=port
    )



import os
import psutil
from surreal.main.ppo_configs import PPOLauncher
from surreal.test_helpers import integration_test


if __name__ == '__main__':
    print('BEGIN DDPG-Gym TEST')
    integration_test('/tmp/surreal/ddpg',
                     os.path.join(os.path.dirname(__file__),
                                  '../surreal/main/ppo_configs.py'),
                     PPOLauncher(),
                     'gym:HalfCheetah-v2')
    print('PASSED')
    self = psutil.Process()
    self.kill()



"""
Checkpoint that supports labeling and performance tracking
"""
import torch
import pickle
import time
import datetime
from benedict import BeneDict
from pkg_resources import parse_version
from collections import OrderedDict
from contextlib import contextmanager
from . import filesys as U


CHEKCPOINT_VERSION = '0.0.1'


class Checkpoint(object):
    """
    Periodically saves attribute of a learner or agent class.
    For PyTorch modules, must extend surreal.utils.pytorch.Module instead of
    the builtin pytorch class.
    One experiment can have multiple Checkpoints with different names. E.g.:
    suppose each agent has different exploration rate annealing. Each agent will
    save to <agent-N>.ckpt and metadata-<agent-N>.yml
    and the central learner will save to learner.ckpt and metadata-learner.yml

    metadata.yml has the following fields:
    - save_counter: how many times "save()" has been called, starts from 1
    - history_ckpt_files: ckpt file names from old to new
    - global_steps: current global_steps counter
    - ckpt: dict of ckpt file names -> info
    """
    # TODO support register custom object serializer method
    def __init__(self, folder,
                 name,
                 *,
                 tracked_obj,
                 tracked_attrs=None,
                 keep_history=1,
                 keep_best=1,
                 mkdir=True):
        """
        Args:
            folder: checkpoint folder path
            name: prefix of the checkpoint
            tracked_obj: arbitrary object whose certain attributes will be tracked.
                typically Learner or Agent class
            tracked_attrs: tracked attribute names of the object, can include
                attrs that point to torch.nn.Module
                set to None if you will restore from existing checkpoint
            keep_history: how many checkpoints to keep, must >= 1
            keep_best: how many best models to keep, sorted by score, must >= 0
            mkdir: True to create intermediate dirs if the path doesn't exist
        """
        self.folder = U.f_expand(folder)
        if mkdir:
            U.f_mkdir(folder)
        self.name = name
        self.tracked_obj = tracked_obj

        if U.f_exists(self.metadata_path()):
            self._load_metadata()
        else:
            # blank-slate checkpoint
            metadata = BeneDict()
            metadata.version = CHEKCPOINT_VERSION
            metadata.save_counter = 0
            metadata.history_ckpt_files = []
            metadata.ckpt = {}
            self._check_tracked_attrs(tracked_attrs)
            metadata.tracked_attrs = tracked_attrs
            assert keep_history >= 1
            metadata.keep_history = keep_history
            assert keep_best >= 0
            metadata.keep_best = keep_best
            metadata.best_ckpt_files = []
            metadata.best_scores = []
            self.metadata = metadata

    def _check_version(self):
        assert 'version' in self.metadata, \
            'version not found in ' + self.metadata_path()
        if parse_version(CHEKCPOINT_VERSION) != parse_version(self.metadata.version):
            raise ValueError('checkpoint version incompatible, please examine '
                             '{} and make sure it is {}'
                             .format(self.metadata_path(), CHEKCPOINT_VERSION))

    def _check_tracked_attrs(self, tracked_attrs):
        ERR_MSG = 'tracked_attrs must be a list of attribute name strings or None'
        if isinstance(tracked_attrs, (list, tuple)):
            for attr_name in tracked_attrs:
                assert isinstance(attr_name, str), ERR_MSG
        else:
            assert tracked_attrs is None, ERR_MSG

    def _load_metadata(self):
        if U.f_exists(self.metadata_path()):
            self.metadata = BeneDict.load_yaml_file(self.metadata_path())
        self._check_version()

    def _restore(self, ckpt_file, check_ckpt_exists):
        """
        helper to be used in self.restore() and self.restore_full_name()
        Returns:
            - ckpt file name if successfully restored
            - None otherwise
        """
        ckpt_path = self._get_path(ckpt_file)
        if not U.f_exists(ckpt_path):
            if check_ckpt_exists:
                raise FileNotFoundError(ckpt_path + ' missing.')
            else:
                return None
        # overwrite attributes on self.tracked_obj
        with open(ckpt_path, 'rb') as fp:
            data = pickle.load(fp)
        for attr_name in self.metadata.tracked_attrs:
            attr_value = getattr(self.tracked_obj, attr_name)
            if isinstance(attr_value, (torch.nn.Module, torch.optim.Optimizer)):
                attr_value.load_state_dict(data[attr_name])
            else:
                setattr(self.tracked_obj, attr_name, data[attr_name])
        return ckpt_path

    @contextmanager
    def _change_folder(self, new_folder):
        """
        Temporary change save folder, useful for loading from other folders
        """
        if new_folder:
            old_folder = self.folder
            new_folder = U.f_expand(new_folder)
            assert U.f_exists(new_folder)
            self.folder = new_folder
            yield
            self.folder = old_folder
        else:
            yield  # noop context

    def restore(self, target,
                mode,
                reload_metadata=True,
                check_ckpt_exists=False,
                restore_folder=None):
        """
        Args:
            target: can be one of the following semantics
              - int: 0 for the last (or best), 1 for the second last (or best), etc.
              - global steps of the ckpt file, the suffix string right before ".ckpt"
            mode: "best" or "history", which group to restore
            reload_metadata: overwrite self.metadata with the metadata.yml file content
            check_ckpt_exists: raise FileNotFoundError if the checkpoint target doesn't exist
            restore_folder: if None, use self.folder, else specify a folder to restore from
                if not None, reload_metadata will be forced to True

        Returns:
            - full ckpt file path if successfully restored
            - None otherwise

        Warnings:
            to restore attrs that point to pytorch Module, you must initialize
            the Module class first, because the checkpoint only saves the
            state_dict, not how to construct your Module
        """
        assert mode in ['best', 'history']
        with self._change_folder(restore_folder):
            if reload_metadata or restore_folder:
                self._load_metadata()
            self._check_version()
            meta = self.metadata
            if isinstance(target, int):
                assert target >= 0, \
                    'target int should start from 0 for the last or best'
                try:
                    if mode == 'best':
                        ckpt_file = meta.best_ckpt_files[target]
                    else:
                        ckpt_file = meta.history_ckpt_files[target]
                except IndexError:
                    if check_ckpt_exists:
                        raise FileNotFoundError(
                            '{} [{}] ckpt file missing'
                            .format(mode.capitalize(), target))
                    else:
                        ckpt_file = '__DOES_NOT_EXIST__'
            else:
                assert '.ckpt' not in target, 'use restore_full_path() instead'
                if mode == 'best':
                    ckpt_file = self.ckpt_name('best-{}'.format(target))
                else:
                    ckpt_file = self.ckpt_name(target)
            return self._restore(ckpt_file, check_ckpt_exists)

    def restore_full_name(self,
                          ckpt_file,
                          check_ckpt_exists=True,
                          restore_folder=None):
        """
        Args:
            ckpt_file: full name of the ckpt_file in self.folder.
            check_ckpt_exists: raise FileNotFoundError if the checkpoint target doesn't exist

        Returns:
            - full ckpt file path if successfully restored
            - None otherwise
        """
        with self._change_folder(restore_folder):
            self._load_metadata()
            return self._restore(ckpt_file, check_ckpt_exists)

    def _get_path(self, fname):
        return U.f_join(self.folder, fname)

    def metadata_name(self):
        return 'metadata.{}.yml'.format(self.name)

    def metadata_path(self):
        return self._get_path(self.metadata_name())

    def ckpt_name(self, suffix):
        return '{}.{}.ckpt'.format(self.name, suffix)

    def _copy_last_to_best(self, suffix):
        U.f_copy(self.ckpt_path(suffix),
                 self.ckpt_path('best-{}'.format(suffix)))

    def ckpt_path(self, suffix):
        return self._get_path(self.ckpt_name(suffix))

    def _save_metadata(self):
        self.metadata.dump_yaml_file(self.metadata_path())

    def _save_ckpt(self, suffix):
        data = OrderedDict()
        assert self.metadata.tracked_attrs is not None, \
            'tracked_attrs must not be None for save(). ' \
            'Did you forget to restore from an existing checkpoint?'
        for attr_name in self.metadata.tracked_attrs:
            attr_value = getattr(self.tracked_obj, attr_name)
            if isinstance(attr_value, (torch.nn.Module, torch.optim.Optimizer)):
                data[attr_name] = attr_value.state_dict()
            else:
                data[attr_name] = attr_value
        with open(self.ckpt_path(suffix), 'wb') as fp:
            pickle.dump(data, fp)

    def save(self, score=None,
             global_steps=None,
             reload_metadata=False,
             **ckpt_info):
        """
        save to `<name>.<global_steps>.ckpt` file

        Args:
            score: float scalar to be compared, for `keep_best`
            global_steps: if None, will use internal save counter
            reload_metadata: True to overwrite self.metadata with disk version
            ckpt_info: additional metadata info for this ckpt
        """
        if reload_metadata:
            self._load_metadata()
        meta = self.metadata
        meta.save_counter += 1
        if global_steps is None:
            global_steps = meta.save_counter
        # save the last step
        suffix = global_steps
        self._save_ckpt(suffix)

        meta.global_steps = global_steps
        # history_ckpt_files count from the newest to the oldest
        meta.history_ckpt_files = [self.ckpt_name(suffix)] + meta.history_ckpt_files
        # delete older history ckpt files.
        ckpt_to_remove = meta.history_ckpt_files[meta.keep_history:]
        for ckpt_name in ckpt_to_remove:
            U.f_remove(self._get_path(ckpt_name))
        del meta.history_ckpt_files[meta.keep_history:]
        # add a ckpt entry to metadata
        ckpt_metadata_entry = {
            'score': score,
            'global_steps': global_steps,
            'save_counter': meta.save_counter,
            'time': time.time(),
            'datetime': str(datetime.datetime.now()),
        }
        ckpt_metadata_entry.update(ckpt_info)
        meta.ckpt[self.ckpt_name(suffix)] = ckpt_metadata_entry
        # save the best-performing weights
        if meta.keep_best > 0:
            assert score is not None, \
                'score cannot be None if keep_best is enabled'
            score_queue = _ScoreQueue(meta.keep_best)
            to_deletes = score_queue.set_queue(
                meta.best_scores, meta.best_ckpt_files)
            best_ckpt_name = self.ckpt_name('best-{}'.format(suffix))
            evict = score_queue.add(score, best_ckpt_name)
            if evict is None or evict[1] != best_ckpt_name:
                # the latest ckpt is not evicted from best queue
                self._copy_last_to_best(suffix)
                meta.ckpt[best_ckpt_name] = ckpt_metadata_entry
            if evict:
                to_deletes.append(evict)
            for _, ckpt_to_delete in to_deletes:
                ckpt_path_to_delete = self._get_path(ckpt_to_delete)
                if U.f_exists(ckpt_path_to_delete):
                    U.f_remove(ckpt_path_to_delete)
                if ckpt_to_delete in meta.ckpt:
                    del meta.ckpt[ckpt_to_delete]
            # print(score_queue.get_scores_filepaths(), best_ckpt_name)
            meta.best_scores, meta.best_ckpt_files = \
                score_queue.get_scores_filepaths()

        self._save_metadata()


class PeriodicCheckpoint(Checkpoint):
    def __init__(self, *args, period, min_interval=0, **kwargs):
        """
        Same signature as Checkpoint
        Args:
            period: Only update when the interval counter % period == 0
            min_interval: Only update when the last update
                          happend min_interval minutes ago
        """
        super().__init__(*args, **kwargs)
        assert period >= 1
        self.period = period
        self._period_counter = 0
        self.min_interval = min_interval
        self.last_update_time = time.time()

    def save(self, *args, **kwargs):
        """
        Same as Checkpoint.save() except that it only runs every
        `period` number of calls.

        Returns:
            True if actually saved
        """
        self._period_counter += 1
        if self._period_counter % self.period == 0:
            if time.time() - self.last_update_time >= self.min_interval:
                super().save(*args, **kwargs)
                self.last_update_time = time.time()
                return True
        return False

    def reset_period(self):
        """
        Set period counter to 0
        """
        self._period_counter = 0


class _ScoreQueue(object):
    """
    Reverse score sorting
    """
    def __init__(self, max_size):
        self._queue = []  # large -> small sorted
        self.max_size = max_size

    def set_queue(self, scores, filepaths):
        "queue of (score, filepath) tuples"
        self._queue = list(zip(scores, filepaths))
        to_delete = self._queue[self.max_size:]
        del self._queue[self.max_size:]
        return to_delete

    def add(self, score, filepath):
        """
        qu = _ScoreQueue(3)
        for i in [1,3,2,4,5,7,3, 8, 2, 2, 1]:
            print(qu.add(i*10, 'path'+str(i)), qu._queue)
        """
        i = len(self._queue)
        while i >= 1:
            sc, _ = self._queue[i - 1]
            if sc > score:
                break
            i -= 1
        self._queue = self._queue[:i] + [(score, filepath)] + self._queue[i:]
        if len(self._queue) > self.max_size:
            to_delete = self._queue[-1]
            del self._queue[-1]
            return to_delete

    def get_scores_filepaths(self):
        "returns score_list, filepath_list"
        return tuple(zip(*self._queue))



from multiprocessing import Process
import os
from caraml.zmq import ZmqProxyThread
import surreal.utils as U


class ShardedReplay(object):
    def __init__(self,
                 replay_class,
                 learner_config,
                 env_config,
                 session_config,):
        """
        Args:
            *_config: passed on to replay
        """
        self.sampler_proxy = None
        self.collector_proxy = None
        self.processes = []

        self.learner_config = learner_config
        self.env_config = env_config
        self.session_config = session_config

        self.replay_class = replay_class

        self.shards = self.learner_config.replay.replay_shards

        self.collector_frontend_port = os.environ['SYMPH_COLLECTOR_FRONTEND_PORT']
        self.collector_backend_port = os.environ['SYMPH_COLLECTOR_BACKEND_PORT']
        self.sampler_frontend_port = os.environ['SYMPH_SAMPLER_FRONTEND_PORT']
        self.sampler_backend_port = os.environ['SYMPH_SAMPLER_BACKEND_PORT']

        self.collector_frontend_add = "tcp://*:{}".format(self.collector_frontend_port)
        self.collector_backend_add = "tcp://*:{}".format(self.collector_backend_port)
        self.sampler_frontend_add = "tcp://*:{}".format(self.sampler_frontend_port)
        self.sampler_backend_add = "tcp://*:{}".format(self.sampler_backend_port)

    def launch(self):

        self.processes = []

        print('Starting {} replay shards'.format(self.shards))
        for i in range(self.shards):
            print('Replay {} starting'.format(i))
            p = Process(target=self.start_replay, args=[i])
            p.start()
            self.processes.append(p)

        self.collector_proxy = ZmqProxyThread(
            in_add=self.collector_frontend_add,
            out_add=self.collector_backend_add,
            pattern='router-dealer')
        self.sampler_proxy = ZmqProxyThread(
            in_add=self.sampler_frontend_add,
            out_add=self.sampler_backend_add,
            pattern='router-dealer')

        self.collector_proxy.start()
        self.sampler_proxy.start()

    def start_replay(self, index):
        replay = self.replay_class(self.learner_config,
                                   self.env_config,
                                   self.session_config,
                                   index=index)
        replay.start_threads()
        replay.join()

    def join(self):
        for i, p in enumerate(self.processes):
            p.join()
            U.report_exitcode(p.exitcode, 'replay-{}'.format(i))
        self.collector_proxy.join()
        self.sampler_proxy.join()



"""
A template class that defines base environment APIs
"""
import surreal.utils as U


class ActionType(U.StringEnum):
    continuous = ()
    discrete = ()


class ObsType(U.StringEnum):
    TODO = ()


class _EnvMeta(type):  # DEPRECATED
    """
    Ensure that env always has `action_spec` and `obs_spec` after __init__
    """
    def __call__(self, *args, **kwargs):
        env = super().__call__(*args, **kwargs)
        # env must have instance vars action_spec and obs_spec
        # they must be dict
        assert hasattr(env, 'action_spec') and isinstance(env.action_spec, dict)
        assert hasattr(env, 'obs_spec') and isinstance(env.obs_spec, dict)
        return env


class Env(object):
    """The main Env class. It encapsulates an environment with
    arbitrary behind-the-scenes dynamics. An environment can be
    partially or fully observed.

    The main API methods that users of this class need to know are:

        step
        reset
        render
        close
        seed

    When implementing an environment, override the following methods
    in your subclass:

        _step
        _reset
        _render
        _close
        _seed

    The methods are accessed publicly as "step", "reset", etc.. The
    non-underscored versions are wrapper methods to which we may add
    functionality over time.
    """
    metadata = {}

    def __new__(cls, *args, **kwargs):
        # We use __new__ since we want the env author to be able to
        # override __init__ without remembering to call super.
        env = super(Env, cls).__new__(cls)
        env._closed = False
        return env

    # Override in SOME subclasses
    def _close(self):
        pass

    # Override in ALL subclasses
    def _step(self, action):
        raise NotImplementedError

    def _reset(self):
        raise NotImplementedError

    def _render(self, *args, **kwargs):
        pass

    def step(self, action):
        """Run one timestep of the environment's dynamics. When end of
        episode is reached, you are responsible for calling `reset()`
        to reset this environment's state.

        Accepts an action and returns a tuple (observation, reward, done, info).

        Args:
            action (object): an action provided by the environment

        Returns:
            observation (object): agent's observation of the current environment
            reward (float) : amount of reward returned after previous action
            done (boolean): whether the episode has ended, in which case further step() calls will return undefined results
            info (dict): contains auxiliary diagnostic information (helpful for debugging, and sometimes learning)
        """
        return self._step(action)

    def reset(self):
        """Resets the state of the environment and returns an initial observation.

        Returns: observation (object): the initial observation of the
            space.
            info (dict)
        """
        return self._reset()

    def render(self, *args, **kwargs):
        """Renders the environment.
        """
        return self._render(*args, **kwargs)

    def close(self):
        """Override _close in your subclass to perform any necessary cleanup.

        Environments will automatically close() themselves when
        garbage collected or when the program exits.

        Also close the rendering threads, if any.
        """
        # _closed will be missing if this instance is still
        # initializing.
        if not hasattr(self, '_closed') or self._closed:
            return
        self._close()
        # If an error occurs before this line, it's possible to
        # end up with double close.
        self._closed = True

    @property
    def unwrapped(self):
        """Completely unwrap this env.

        Returns:
            gym.Env: The base non-wrapped gym.Env instance
        """
        return self

    def __del__(self):
        self.close()

    def __str__(self):
        return '<{}>'.format(type(self).__name__)



import torch
import torch.nn as nn
import torch.nn.functional as F
import surreal.utils as U


class TorsoBuilder(nn.Module):

    def __init__(self,
                 input_spec,
                 conv_spec=None,
                 mlp_spec=None,
                 rnn_spec=None):

        super(TorsoBuilder, self).__init__()

        if conv_spec:
            assert len(input_spec['dims']) == 3, 'torso requires 3-dimensional inputs to conv layers.'
        elif mlp_spec:
            assert len(input_spec['dims']) == 1, 'torso requires flat inputs to fc layers.'
        elif rnn_spec:
            raise NotImplementedError
        else:
            raise ValueError('torso must contain at least one layer.')

        self._conv_spec = conv_spec
        self._mlp_spec = mlp_spec
        self._rnn_spec = rnn_spec

        self._input_spec = input_spec
        self.conv_module = None
        self.mlp_module = None
        self.rnn_module = None

    def forward(self, obs):

        # verify obs shapes
        for k, d in enumerate(self._input_spec['dims']):
            assert obs.size(k+1) == d

        out = obs

        if self._conv_spec:
            if not self.conv_module:
                self.conv_module = build_conv_module(self._input_spec,
                                                     **self._conv_spec)
            out = self.conv_module(out)

        if self._mlp_spec:
            # lazy evaluation for shape inference
            out = out.view(out.size(0), -1)
            if not self.mlp_module:
                mlp_input_spec = {'dims': [out.size(1)]}
                self.mlp_module = build_mlp_module(mlp_input_spec,
                                                   **self._mlp_spec)
            out = self.mlp_module(out)

        return out


def build_conv_module(input_spec,
                      out_channels,
                      kernel_sizes,
                      strides=None,
                      paddings=None,
                      dilations=None,
                      use_batch_norm=False):

    layers = []
    in_channels = input_spec['dims'][0]

    for i in range(len(kernel_sizes)):

        stride = strides[i] if strides else 1
        padding = paddings[i] if paddings else 0
        dilation = dilations[i] if dilations else 1

        conv2d = nn.Conv2d(in_channels=in_channels,
                           out_channels=out_channels[i],
                           kernel_size=kernel_sizes[i],
                           stride=stride,
                           padding=padding,
                           dilation=dilation)

        in_channels = out_channels[i]

        if use_batch_norm:
            layers += [conv2d, nn.BatchNorm2d(in_channels), nn.ReLU(inplace=True)]
        else:
            layers += [conv2d, nn.ReLU(inplace=True)]

    layers = nn.Sequential(*layers)
    return layers


def build_mlp_module(input_spec,
                     sizes,
                     use_dropout=False):

    assert len(input_spec['dims']) == 1, 'MLP requires flat inputs.'
    assert len(sizes) > 0, 'MLP must have at least one fc layer.'

    layers = []
    prev_dim = input_spec['dims'][0]
    for i in range(len(sizes)):
        layers += [nn.Linear(prev_dim , sizes[i]), nn.ReLU(inplace=True)]
        if use_dropout:
            layers += [nn.Dropout()]
        prev_dim = sizes[i]

    layers = nn.Sequential(*layers)
    return layers



import os
import re
import shlex
import subprocess
from copy import copy
from pathlib import Path
from symphony.commandline import SymphonyParser
from symphony.engine import SymphonyConfig, Cluster
from symphony.kube import GKEDispatcher
from symphony.addons import clean_images
from benedict import BeneDict
from surreal.launch import (
    SurrealDockerBuilder,
    CommandGenerator,
    setup_network,
)
import surreal.utils as U


def _merge_setting_dictionaries(customize, base):
    di = copy(base)
    for key in di:
        if isinstance(di[key], dict):
            if key in customize:
                di[key] = _merge_setting_dictionaries(customize[key], di[key])
        else:
            if key in customize and customize[key] is not None:
                di[key] = customize[key]
    return di


class KurrealParser(SymphonyParser):
    def create_cluster(self):
        return Cluster.new('kube')

    def setup(self):
        super().setup()
        self.docker_build_settings = {}
        self.config = BeneDict()
        self.load_config()
        self._setup_create()
        self._setup_tensorboard()
        self._setup_docker_clean()

        # Secondary nfs related support
        self._setup_get_videos()
        self._setup_get_config()
        self._setup_get_tensorboard()

    def load_config(self):
        surreal_yml_path = U.get_config_file()
        if not U.f_exists(surreal_yml_path):
            raise ValueError('Cannot find surreal config file at {}'
                             .format(surreal_yml_path))
        self.config = BeneDict.load_yaml_file(surreal_yml_path)
        SymphonyConfig().set_username(self.username)
        SymphonyConfig().set_experiment_folder(self.folder)

        if 'docker_build_settings' in self.config:
            for setting in self.config['docker_build_settings']:
                self.docker_build_settings[setting['name']] = setting

    @property
    def folder(self):
        return U.f_expand(self.config.kube_metadata_folder)

    @property
    def username(self):
        assert 'username' in self.config, 'must specify username in ~/.surreal.yml'
        return self.config.username

    def _setup_get_videos(self):
        parser = self.add_subparser('get-videos', aliases=['gv'])
        parser.add_argument('experiment_names', nargs='*', type=str, metavar='experiment_name',
                            help='experiments to retrieve videos for, '
                            'none to retrieve your own running experiments')
        parser.add_argument('--last', type=int, default=5, metavar='last_n_videos',
                            help='Number of most recent videos, -1 to get all')
        parser.add_argument('--save_folder', type=str, default='.',
                            help='save_videos in [save_folder]/experiment_name')

    def _setup_get_config(self):
        parser = self.add_subparser('get-config', aliases=['gc'])
        parser.add_argument('experiment_name', type=str,
                            help='experiments to retrieve videos for, '
                                 'none to retrieve your own running experiments')
        parser.add_argument('-o', '--output-file', type=str,
                            help='save remote config to a specified local file path')

    def _setup_get_tensorboard(self):
        parser = self.add_subparser('get-tensorboard', aliases=['gt'])
        parser.add_argument('experiment_name', type=str,
                            help='experiments to retrieve tensorboard for, '
                                 'none to retrieve your own running experiments')
        parser.add_argument('-s', '--subfolder', type=str, default='',
                            help='retrieve only a subfolder under the "tensorboard" folder. '
                                 'currently valid folders are agent, eval, learner, replay')
        parser.add_argument('-o', '--output-folder', type=str,
                            help='save remote TB folder to a specified local folder path')

    def _setup_docker_clean(self):
        parser = self.add_subparser('docker-clean', aliases=['dc'])

    def _setup_tensorboard(self):
        parser = self.add_subparser('tensorboard', aliases=['tb'])
        self._add_experiment_name(parser, required=False, positional=True)
        parser.set_defaults(service_name='tensorboard')
        parser.add_argument(
            '-u', '--url-only',
            action='store_true',
            help='only show the URL without opening the browser.'
        )

    def _setup_create(self):
        parser = self.add_subparser('create', aliases=['c'])
        parser.add_argument(
            'setting_name',
            type=str,
            help='the setting in .surreal.yml that specifies how an'
                 'experiment should be run')
        self._add_experiment_name(parser)
        parser.add_argument(
            '--algorithm',
            type=str,
            help='ddpg / ppo or the'
                 'location of algorithm python script **in the docker '
                 'container**'
        )
        parser.add_argument(
            '--num_agents',
            type=int,
            default=None,
            help='number of agent pods to run in parallel.'
        )
        parser.add_argument(
            '--num_evals',
            type=int,
            default=None,
            help='number of eval pods to run in parallel.'
        )
        parser.add_argument(
            '--agent-batch',
            type=int,
            default=None,
            help='put how many agent on each agent pod'
        )
        parser.add_argument(
            '--eval-batch',
            type=int,
            default=None,
            help='put how many eval on each eval pod'
        )
        parser.add_argument(
            '--env',
            type=str,
            default=None,
            help='What environment to run'
        )

        parser.add_argument(
            '-f', '--force',
            action='store_true',
            help='force overwrite an existing kurreal.yml file '
                 'if its experiment folder already exists.'
        )
        self._add_dry_run(parser)

    # ==================== helpers ====================
    def _add_dry_run(self, parser):
        parser.add_argument(
            '-dr', '--dry-run',
            action='store_true',
            help='print the kubectl command without actually executing it.'
        )

    def _process_experiment_name(self, experiment_name):
        """
        experiment_name will be used as DNS, so must not have underscore or dot
        """
        new_name = experiment_name.lower().replace('.', '-').replace('_', '-')
        if new_name != experiment_name:
            print('experiment name string has been fixed: {} -> {}'
                  .format(experiment_name, new_name))
        return new_name

    SUPPORTED_MODES = ['basic']

    def action_create(self, args):
        """
            Spin up a multi-node distributed Surreal experiment.
            Put any command line args that pass to the config script after "--"
        """
        setting_name = args.setting_name

        if not setting_name in self.config.creation_settings:
            raise KeyError('Cannot find setting {}'.format(setting_name))
        settings = self.config.creation_settings[setting_name]
        mode = settings['mode']
        if mode not in self.SUPPORTED_MODES:
            raise ValueError('Unknown mode {}'.format(mode) +
                             'available options are : {}'.format(
                                 ', '.join(self.SUPPORTED_MODES)
                             ))
        if mode == 'basic':
            self.create_basic(
                settings=settings,
                experiment_name=args.experiment_name,
                algorithm_args=args.remainder,
                input_args=vars(args),
                force=args.force,
                dry_run=args.dry_run,
                )

    def _find_executable(self, name):
        """
            Finds the .py file corresponding to the algorithm specified

        Args:
            name: ddpg / ppo / <path in container to compatible .py file>
        """
        if name == 'ddpg':
            return 'surreal-ddpg'
        elif name == 'ppo':
            return 'surreal-ppo'
        else:
            return name

    DEFAULT_SETTING_BASIC = {
        'algorithm': 'ddpg',
        'num_agents': 2,
        'num_evals': 1,
        'agent_batch': 1,
        'eval_batch': 1,
        'restore_folder': None,
        'env': 'gym:HalfCheetah-v2',
        'agent': {
            'image': 'surrealai/surreal-nvidia:v0.1',
            'build_image': None,
            'scheduling': {
                'assign_to': 'node_pool',
                'node_pool_name': None,
                'cpu': None,
                'memory_m': None,
                'gpu_type': None,
                'gpu_count': None,
            }
        },
        'nonagent': {
            'image': 'surrealai/surreal-nvidia:v0.1',
            'build_image': None,
            'scheduling': {
                'assign_to': 'node_pool',
                'node_pool_name': None,
                'cpu': None,
                'memory_m': None,
                'gpu_type': None,
                'gpu_count': None,
            }
        },
    }

    def create_basic(self, *,
                     settings,
                     experiment_name,
                     algorithm_args,
                     input_args,
                     force,
                     dry_run):
        settings = _merge_setting_dictionaries(settings,
                                               self.DEFAULT_SETTING_BASIC)
        settings = _merge_setting_dictionaries(input_args, settings)
        settings = BeneDict(settings)

        cluster = self.create_cluster()
        if 'mount_secrets' in self.config:
            secrets = self.config.mount_secrets
        else:
            secrets = None
        exp = cluster.new_experiment(experiment_name, secrets=secrets)

        image_builder = SurrealDockerBuilder(
            build_settings=self.docker_build_settings,
            images_requested={
                'agent': {
                    'identifier': settings.agent.image,
                    'build_config': settings.agent.build_image
                },
                'nonagent': {
                    'identifier': settings.nonagent.image,
                    'build_config': settings.nonagent.build_image
                },
            },
            tag=exp.name,
            push=True)
        agent_image = image_builder.images_provided['agent']
        nonagent_image = image_builder.images_provided['nonagent']
        # defer to build last, so we don't build unless everything passes

        algorithm_args += [
            "--num-agents",
            str(settings.num_agents * settings.agent_batch),
            ]
        # TODO: restore_functionalities
        if settings.restore_folder is not None:
            algorithm_args += ["--restore_folder",
                               shlex.quote(settings.restore_folder)]
        experiment_folder = self.get_remote_experiment_folder(exp.name)
        algorithm_args += ["--experiment-folder",
                           str(experiment_folder)]
        algorithm_args += ["--env", str(settings.env)]
        algorithm_args += ["--agent-batch", str(settings.agent_batch)]
        algorithm_args += ["--eval-batch", str(settings.eval_batch)]
        executable = self._find_executable(settings.algorithm)
        cmd_gen = CommandGenerator(
            num_agents=settings.num_agents,
            num_evals=settings.num_evals,
            executable=executable,
            config_commands=algorithm_args)

        nonagent = exp.new_process_group('nonagent')
        learner = nonagent.new_process(
            'learner',
            container_image=nonagent_image,
            args=[cmd_gen.get_command('learner')])
        # Because learner and everything are bundled together

        # json_path = 'cluster_definition.tf.json'  # always use slash
        # filepath = pkg_resources.resource_filename(__name__, json_path)
        json_path = self.config.cluster_definition
        dispatcher = GKEDispatcher(json_path)
        # We only need to claim resources for learner
        dispatcher.assign_to(learner,
                             process_group=nonagent,
                             **settings.nonagent.scheduling)
        # For dm_control
        learner.set_env('DISABLE_MUJOCO_RENDERING', "1")

        replay = nonagent.new_process(
            'replay',
            container_image=nonagent_image,
            args=[cmd_gen.get_command('replay')])

        ps = nonagent.new_process(
            'ps',
            container_image=nonagent_image,
            args=[cmd_gen.get_command('ps')])

        tensorboard = nonagent.new_process(
            'tensorboard',
            container_image=nonagent_image,
            args=[cmd_gen.get_command('tensorboard')])

        tensorplex = nonagent.new_process(
            'tensorplex',
            container_image=nonagent_image,
            args=[cmd_gen.get_command('tensorplex')])

        loggerplex = nonagent.new_process(
            'loggerplex',
            container_image=nonagent_image,
            args=[cmd_gen.get_command('loggerplex')])
        nonagent.image_pull_policy('Always')

        agents = []
        for i in range(settings.num_agents):
            if settings.agent_batch == 1:
                agent_name = 'agent-{}'.format(i)
            else:
                agent_name = 'agents-{}'.format(i)
            agent = exp.new_process(
                agent_name,
                container_image=agent_image,
                args=[cmd_gen.get_command(agent_name)])

            agent.image_pull_policy('Always')
            # We only need to claim resources for learner
            dispatcher.assign_to(agent,
                                 **settings.agent.scheduling)
            agents.append(agent)

        evals = []
        for i in range(settings.num_evals):
            if settings.eval_batch == 1:
                eval_name = 'eval-{}'.format(i)
            else:
                eval_name = 'evals-{}'.format(i)
            eval_p = exp.new_process(
                eval_name,
                container_image=agent_image,
                args=[cmd_gen.get_command(eval_name)])
            dispatcher.assign_to(eval_p, **settings.agent.scheduling)
            eval_p.image_pull_policy('Always')

            evals.append(eval_p)

        setup_network(agents=agents,
                      evals=evals,
                      learner=learner,
                      replay=replay,
                      ps=ps,
                      tensorboard=tensorboard,
                      tensorplex=tensorplex,
                      loggerplex=loggerplex)

        if 'nfs' in self.config:
            print('NFS mounted')
            nfs_server = self.config.nfs.servername
            nfs_server_path = self.config.nfs.fs_location
            nfs_mount_path = self.config.nfs.mount_path
            for proc in exp.list_all_processes():
                proc.mount_nfs(server=nfs_server,
                               path=nfs_server_path,
                               mount_path=nfs_mount_path)
        else:
            print('NFS not mounted')

        image_builder.build()
        cluster.launch(exp, force=force, dry_run=dry_run)

    def get_remote_experiment_folder(self, experiment_name):
        """
            Actual experiment folder will be
            <mount_path>/<root_subfolder>/<experiment_name>/
        """
        # DON'T use U.f_join because we don't want to expand the path locally
        directory = self.config.kube_results_folder
        return os.path.join(directory, experiment_name)

    def action_tensorboard(self, args):
        self.action_visit(args)

    def action_docker_clean(self, args):
        """
            Cleans all docker images used to create experiments
        """
        potential_images = []
        for name, settings in self.config.creation_settings.items():
            if settings.mode == 'basic':
                if 'agent' in settings and 'image' in settings.agent:
                    potential_images.append(settings.agent.image)
                if 'nonagent' in settings and 'image' in settings.nonagent:
                    potential_images.append(settings.nonagent.image)
            else:
                raise ValueError('Unsupported creation setting mode {}'
                                 .format(settings.mode) + ' for creation'
                                 + ' setting ' + name)

        potential_images = list(set(potential_images))
        potential_images = [x for x in potential_images if ':' not in x]
        images_to_clean = ['{}:*'.format(x) for x in potential_images]
        clean_images(images_to_clean)

    def action_get_videos(self, args):
        self._check_nfs_retrieve_settings()
        remote_folder = Path(self.config.nfs.results_folder)
        todos = []
        if len(args.experiment_names) == 0:
            experiments = self.cluster.list_experiments()
            for experiment in experiments:
                if re.match(self.username, experiment):
                    todos.append(experiment)
        else:
            todos = args.experiment_names

        print('Fetching videos for:')
        print('\n'.join(['\t' + x for x in todos]))

        save_folder = os.path.expanduser(args.save_folder)
        save_last = args.last

        for experiment_name in todos:
            self._get_video_for_experiment(
                experiment_name, save_folder, remote_folder, save_last
            )

    def _get_video_for_experiment(self,
                                  experiment_name,
                                  save_folder,
                                  remote_folder,
                                  save_last=-1):
        # Find remote path
        remote_folder = remote_folder / experiment_name / 'videos'
        # parse existing files
        results = self._gcloud_nfs_exec('ls -1 {}'.format(remote_folder))
        video_files = results.strip().split('\n')
        video_episodes = [x[len('video_eps_'):] for x in video_files]
        video_episodes = [int(x[:len(x) - len('.mp4')]) for x in video_episodes]
        video_episodes = sorted(video_episodes, reverse=True)
        if save_last > 0:
            save_last = min(save_last, len(video_episodes))
            video_episodes = video_episodes[:save_last]
        filenames = ['video_eps_{}.mp4'.format(x) for x in video_episodes]
        # local path
        local_folder = Path(os.path.expanduser(save_folder)) / experiment_name
        local_folder.mkdir(exist_ok=True, parents=True)
        # download
        for filename in filenames:
            print('$> get {}'.format(str(remote_folder / filename)))
            self._gcloud_download(remote_folder / filename, local_folder / filename)
        return filenames

    def action_get_config(self, args):
        """
        Download remote config.yml in the experiment folder
        """
        self._check_nfs_retrieve_settings()
        remote_folder = Path(self.config.nfs.results_folder)
        if args.output_file:
            output_file = args.output_file
        else:
            output_file = 'config.yml'
        experiment_name = args.experiment_name
        remote_config_file = (remote_folder / experiment_name / 'config.yml')
        print('Downloading', remote_config_file)
        self._gcloud_download(remote_config_file, output_file)

    def action_get_tensorboard(self, args):
        """
        Download remote config.yml in the experiment folder
        """
        self._check_nfs_retrieve_settings()
        remote_folder = Path(self.config.nfs.results_folder)
        experiment_name = args.experiment_name
        if args.output_folder:
            output_folder = args.output_folder
        else:
            output_folder = experiment_name
        remote_tb_folder = (remote_folder / experiment_name
                            / 'tensorboard' / args.subfolder)
        print('Downloading', remote_tb_folder)
        self._gcloud_download(remote_tb_folder, output_folder)

    def _gcloud_download(self, remote_path, local_path):
        cmd = "gcloud compute scp --recurse {}:'{}' '{}'".format(
            self.config.nfs.servername, remote_path, local_path
        )
        os.system(cmd)

    def _gcloud_nfs_exec(self, command):
        return subprocess.check_output(
            "gcloud compute ssh {} -- '{}'".format(self.config.nfs.servername, command),
            shell=True
        ).decode('utf-8').replace('\r\n', '\n')


    def _check_nfs_retrieve_settings(self):
        if 'nfs' not in self.config:
            raise ValueError('nfs field not found in .surreal.yml, aborting')
        if 'servername' not in self.config.nfs:
            raise ValueError('nfs:servername field not found in .surreal.yml, aborting')
        if 'results_folder' not in self.config.nfs:
            raise ValueError('nfs:results_folder not found .surreal.yml, aborting')


def main():
    KurrealParser().main()


if __name__ == '__main__':
    main()



"""
Agent side.
Send experience chunks (buffered) to Replay node.
"""
import surreal.utils as U
from surreal.session import PeriodicTracker
from caraml.zmq import ZmqSender


class ExpBuffer(object):
    """
        Temporarily holds and deduplicates experience
    """
    def __init__(self):
        self.exp_list = []  # list of exp dicts
        self.ob_storage = {}

    def add(self, hash_dict, nonhash_dict):
        """
        Args:
            hash_dict: {obs_hash: [ .. can be nested .. ]}
            nonhash_dict: {reward: -1.2, done: True, ...}
        """
        U.assert_type(hash_dict, dict)
        U.assert_type(nonhash_dict, dict)
        exp = {}
        for key, values in hash_dict.items():
            assert not key.endswith('_hash'), 'do not manually append `_hash`'
            exp[key + '_hash'] = self._hash_nested(values)
        exp.update(nonhash_dict)
        self.exp_list.append(exp)

    def flush(self):
        """
        Serialized all currenct content of the buffer into binary

        Returns:
            binary data of (exp_list, ob_storage)
        """
        binary = U.serialize((self.exp_list, self.ob_storage))
        self.exp_list = []
        self.ob_storage = {}
        return binary

    def _hash_nested(self, values):
        if isinstance(values, list):
            return [self._hash_nested(v) for v in values]
        if isinstance(values, tuple):
            return tuple([self._hash_nested(v) for v in values])
        elif isinstance(values, dict):
            return {k: self._hash_nested(v) for k, v in values.items()}
        elif values is None:
            return None
        else:  # values is a single object
            obj = values
            hsh = U.pyobj_hash(obj)
            if hsh not in self.ob_storage:
                self.ob_storage[hsh] = obj
            return hsh


class ExpSender(object):
    """
    `send()` logic can be overwritten to support
    more complicated agent experiences,
    such as multiagent, self-play, etc.
    """
    def __init__(self, *,
                 host,
                 port,
                 flush_iteration):
        """
        Args:
            flush_iteration: how many send() calls before we flush the buffer
        """
        U.assert_type(flush_iteration, int)
        self._client = ZmqSender(host=host,
                                 port=port)
        self._exp_buffer = ExpBuffer()
        self._flush_tracker = PeriodicTracker(flush_iteration)

    def send(self, hash_dict, nonhash_dict):
        """
        Args:
            hash_dict: Large/Heavy data that should be deduplicated
                       by the caching mekanism
            nonhash_dict: Small data that we can afford to keep copies of
        """
        self._exp_buffer.add(
            hash_dict=hash_dict,
            nonhash_dict=nonhash_dict,
        )
        if self._flush_tracker.track_increment():
            exp_binary = self._exp_buffer.flush()
            self._client.send(exp_binary)
            return U.binary_hash(exp_binary)
        else:
            return None



"""
Aggregate experience tuple into pytorch-ready tensors
"""
import numpy as np
import copy
import collections
import torch
import surreal.utils as U
from surreal.env import ActionType

class FrameStackPreprocessor():
    """
        Accepts experience sent by SSAR experience senders
    """

    def __init__(self, frame_stacks):
        self.frame_stacks = frame_stacks
        self.printed = False

    def preprocess_obs(self, obs):
        # We must copy here because the experience sender should send the non-stacked version
        if 'pixel' in obs:
            for key in obs['pixel']:
                obs['pixel'][key] = np.concatenate(obs['pixel'][key], axis=0)
                assert len(obs['pixel'][key].shape) == 3

    def preprocess_list(self, exp_list):
        for exp in exp_list:  # dict
            for obs in (exp['obs'][0], exp['obs'][1]):
                self.preprocess_obs(obs)
        return exp_list

class SSARAggregator():
    """
        Accepts experience sent by SSAR experience senders
        aggregate() returns float arrays:
        {
            obs = batch_size * observation
            obs_next = batch_size * next_observation
            actions = batch_size * actions,
            rewards = batch_size * 1,
            dones = batch_size * 1,
        }
    """
    def __init__(self, obs_spec, action_spec):
        U.assert_type(obs_spec, dict)
        U.assert_type(action_spec, dict)
        self.action_type = ActionType[action_spec['type']]
        self.action_spec = action_spec
        self.obs_spec = obs_spec

    def aggregate(self, exp_list):
        # TODO add support for more diverse obs_spec and action_spec
        """

        Args:
            exp_list:
        
        Returns:
            aggregated experience
        """

        obs0, actions, rewards, obs1, dones = (
            collections.OrderedDict(), [], [], collections.OrderedDict(), [])
        i = 0
        for exp in exp_list:  # dict
            i += 1
            for modality in exp['obs'][0]:
                if modality not in obs0:
                    obs0[modality] = collections.OrderedDict()
                for key in exp['obs'][0][modality]:
                    if key not in obs0[modality]:
                        obs0[modality][key] = []
                    obs0[modality][key].append(np.array(exp['obs'][0][modality][key], copy=False))
            actions.append(exp['action'])
            rewards.append(exp['reward'])
            for modality in exp['obs'][1]:
                if modality not in obs1:
                    obs1[modality] = collections.OrderedDict()
                for key in exp['obs'][1][modality]:
                    if key not in obs1[modality]:
                        obs1[modality][key] = []
                    obs1[modality][key].append(np.array(exp['obs'][1][modality][key], copy=False))
            dones.append(float(exp['done']))
        if self.action_type == ActionType.continuous:
            actions = np.array(actions, dtype=np.float32)
        elif self.action_type == ActionType.discrete:
            actions = np.array(actions, dtype=np.int32)
        else:
            raise NotImplementedError('action_spec unsupported '+str(self.action_spec))

        for obs in obs0, obs1:
            for modality in obs:
                for key in obs[modality]:
                    obs[modality][key] = np.array(obs[modality][key])

        return {
            'obs': obs0,
            'obs_next': obs1,
            'actions': np.array(actions),
            'rewards': np.expand_dims(rewards, axis=1),
            'dones': np.expand_dims(dones, axis=1),
        }


class MultistepAggregatorWithInfo():
    """
        Accepts input by ExpSenderWrapperMultiStepMovingWindowWithInfo
        aggregate() returns float Tensors
        {
            obs = batch_size * n_step * observation
            next_obs = batch_size * 1 * next_observation
            actions = batch_size * n_step * actions,
            rewards = batch_size * n_step,
            persistent_infos = list of batched FloatTensor action info per step,
            onetime_infos = FloatTensor of action info for first step of batch,
            dones = batch_size * n_step,
        }

        This class of aggregator is used in companion with exp sender.
        ExpSenderWrapperMultiStepMovingWindowWithInfo. The key return values are
        a list of persistent infos, and a FloatTensor of onetime info that are 
        auxiliary information returned from the agent. The main difference is
        that persistent infos are required for every step, for instance the probability
        distribution for policy, whereas onetime infos are needed only for first or 
        last of the steps, for example the LSTM hidden/cell states.

        This aggregator should be used when the agent is required to 
        expose attributes to the learner. For instance, PPO needs to use this
        aggregator to send over the action probability distribution and optional
        RNN hidden states.

        This attribute will be returned as a list of batched FloatTensors. In
        the case of PPO without RNN policy, the action_infos attribute is of
        form:
            persistent info: [FloatTensor of shape (batch_size, 2 * act_dim)]
            onetime infos: []

        In case with RNN:
            persistent info: [FloatTensor of shape (batch, horizon, 2 * act_dim)]
            onetime infos: [LSTM hidden state, LSTM cell state]

    """
    def __init__(self, obs_spec, action_spec):
        U.assert_type(obs_spec, dict)
        U.assert_type(action_spec, dict)
        self.action_type = ActionType[action_spec['type']]
        self.action_spec = action_spec
        self.obs_spec = obs_spec

    def aggregate(self, exp_list):
        '''
            Aggregates a list of subtrajectory dictionaries into dictionary of 
            batched sub-trajectory
            Args:
                exp_list: list of dictionaries that are sub-trajectory attribute
            Returns:
                dict of tensorized subtrajectory information
        '''
        observations, next_obs, actions, rewards, dones, persistent_infos, onetime_infos = [], [], [], [], [], [], []
        for exp in exp_list:
            action_n_step, reward_n_step, done_n_step = self._stack_n_step_experience(exp)
            actions.append(action_n_step)
            rewards.append(reward_n_step)
            dones.append(done_n_step)
            observations.append(exp['obs'])
            next_obs.append([exp['obs_next']])

        observations = self._batch_obs(observations)
        next_obs = self._batch_obs(next_obs)

        if self.action_type == ActionType.discrete:
            actions = np.expand_dims(2).astype('int64')
        elif self.action_type is not ActionType.continuous:
            raise NotImplementedError('action_spec unsupported '+str(self.action_spec))

        onetime_infos, persistent_infos = self._gather_action_infos(exp_list)
        return {'obs': observations,
                'obs_next': next_obs,
                'actions': np.stack(actions),
                'rewards': np.stack(rewards),
                'persistent_infos': persistent_infos,
                'onetime_infos': onetime_infos,
                'dones': np.stack(dones).astype('float32')}

    def _batch_obs(self, traj_list):
        '''
            Helper function that batches a list of observation (nested dictionary)
            into one whole dictionary of batched observation
            Args:
                traj_list: list of nested observation
        '''
        batched_obs = {}
        for modality in self.obs_spec.keys():
            batched_obs[modality] = {}
            for key in self.obs_spec[modality].keys():
                batched_obs[modality][key] = []
                for exp in traj_list: # exp ~= ex['obs']
                    n_step_obs = []
                    for obs in exp: 
                        n_step_obs.append(obs[modality][key])
                    n_step_obs = np.stack(n_step_obs)
                    batched_obs[modality][key].append(n_step_obs)
                batched_obs[modality][key] = np.stack(batched_obs[modality][key])
        return batched_obs
        
    def _stack_n_step_experience(self, experience):
        """
            Stacks n steps into single numpy arrays
            Args:
                experience: (type: dictionary) subtrajectory information
            Returns:
                observations: stacked numpy array for observations in dim 0
                actions: stacked numpy array for actions in dim 0
                rewards: stacked numpy array for rewards in dim 0
                dones: stacked numpy array for termination flag in dim 0
        """
        actions = np.stack(experience['actions'])
        rewards = np.array(experience['rewards'])
        dones = np.array(experience['dones'])
        return actions, rewards, dones

    def _gather_action_infos(self, exp_list):
        """
            Gathers corresponding action informations from partial trajectories
            Args:
                experience: (type: dictionary) subtrajectory information
            Returns:
                persistent_infos: list of batched FloatTensors
                onetime_infos: one batched FloatTensor
        """
        persistent_infos, onetime_infos = None, None

        exists_ones = (len(exp_list[0]['onetime_infos']) > 0)
        exists_pers = (len(exp_list[0]['persistent_infos'][0]) > 0)

        if exists_ones:
            onetime_infos = []
            for _ in range(len(exp_list[0]['onetime_infos'])):
                onetime_infos.append([])
        if exists_pers:
            persistent_infos = []
            for _ in range(len(exp_list[0]['persistent_infos'][0])):
                persistent_infos.append([])

        for exp in exp_list:
            if exists_ones:
                for i in range(len(onetime_infos)):
                    onetime_infos[i].append(exp['onetime_infos'][i])
            if exists_pers:
                for i in range(len(persistent_infos)):
                    one_exp_info = []
                    for info_list in exp['persistent_infos']:
                        one_exp_info.append(info_list[i])
                    persistent_infos[i].append(np.stack(one_exp_info))

        if exists_ones:
            onetime_infos = [np.stack(info) for info in onetime_infos]
        if exists_pers:
            persistent_infos = [np.asarray(infos) for infos in persistent_infos]

        return onetime_infos, persistent_infos

class NstepReturnAggregator():
    """
        Accepts input by ExpSenderWrapperMultiStepMovingWindow
        Sums over the moving window for bootstrap n_step return

        {
            obs = batch_size * observation
            obs_next = batch_size * next_observation
            actions = batch_size * actions,
            rewards = batch_size,
            dones = batch_size,
        }
    """
    def __init__(self, obs_spec, action_spec, gamma):
        U.assert_type(obs_spec, dict)
        U.assert_type(action_spec, dict)
        self.action_type = ActionType[action_spec['type']]
        self.action_spec = action_spec
        self.obs_spec = obs_spec
        self.gamma = gamma
    
    def aggregate(self, exp_list):
        # TODO add support for more diverse obs_spec and action_spec
        """

        Args:
            exp_list:
        
        Returns:
            aggregated experience
        """
        
        obs0, actions, rewards, obs1, dones, num_steps = (
            collections.defaultdict(list), [], [], collections.defaultdict(list), [], [])
        for exp in exp_list:  # dict
            n_step = exp['n_step']
            num_steps.append(n_step)
            for key in exp['obs_arr'][0]:
                obs0[key].append(np.array(exp['obs_arr'][0][key], copy=False))
            actions.append(exp['action_arr'][0])
            cum_reward = 0
            for i, r in enumerate(exp['reward_arr']):
                cum_reward += pow(self.gamma, i) * r
            rewards.append(cum_reward)
            for key in exp['obs_next']:
                obs1[key].append(np.array(exp['obs_next'][key], copy=False))
            dones.append(float(exp['done_arr'][n_step - 1]))
        for obs in [obs0, obs1]:
            for key in obs:
                obs[key] = np.array(obs[key])

        # TODO: convert action to appropriate numpy array
        if self.action_type == ActionType.continuous:
            actions = np.array(actions, dtype=np.float32)
        elif self.action_type == ActionType.discrete:
            actions = np.array(actions, dtype=np.int32)
        else:
            raise NotImplementedError('action_spec unsupported '+str(self.action_spec))

        return {
            'obs': dict(obs0),
            'obs_next': dict(obs1),
            'actions': np.array(actions),
            'rewards': np.expand_dims(np.array(rewards), axis=1),
            'dones': np.expand_dims(np.array(dones), axis=1),
            'num_steps': np.expand_dims(np.array(num_steps), axis=1),
        }



import torch
import torch.nn as nn
import surreal.utils as U
import numpy as np
import torchx.nn as nnx

class ZFilter(nnx.Module):
    """
        Keeps historical average and std of inputs
        Whitens data and clamps to +/- 5 std
        Attributes:
            in_size: state dimension
                required from input
            eps: tolerance value for computing Z-filter (whitening)
                default to 10
            running_sum: running sum of all previous states
                (Note, type is torch.cuda.FloatTensor or torch.FloatTensor)
            running_sumsq: sum of square of all previous states
                (Note, type is torch.cuda.FloatTensor or torch.FloatTensor)
            count: number of experiences accumulated
                (Note, type is torch.cuda.FloatTensor or torch.FloatTensor)
    """
    def __init__(self, obs_spec, eps=1e-5):
        """
        Constructor for ZFilter class
        Args:
            obs_spec: nested dictionary of observation space spec. see doc
            eps: tolerance value for computing Z-filter (whitening)
        """
        super(ZFilter, self).__init__()
        self.eps = eps
        self.obs_spec = obs_spec

        in_size = 0
        for key in self.obs_spec['low_dim'].keys():
            in_size += self.obs_spec['low_dim'][key][0]
        self.in_size = in_size

        # Keep some buffers for doing whitening. 
        self.register_buffer('running_sum', torch.zeros(in_size))
        self.register_buffer('running_sumsq', eps * torch.ones(in_size))
        self.register_buffer('count', torch.tensor([eps], dtype=torch.float32))

    def z_update(self, x):
        """
            Count x into historical average, updates running sums and count
            Accepts batched input
            Args:
                x: input tensor to be kept in record. 
        """
        if x is None: return

        # only called in learner, so we can assume it has the correct type
        if len(x.size()) == 3: x = x.view(-1, self.in_size)
        self.running_sum += torch.sum(x, dim=0)
        self.running_sumsq += torch.sum(x * x, dim=0)
        self.count += float(len(x))

    def forward(self, inputs):
        '''
            Whiten observation (inputs) to have zero-mean, unit variance.
            Also clamps output to be within 5 standard deviations
            Args:
                inputs -- batched observation input. batch size at dim 0
            Returns:
                0 mean std 1 weightened batch of observation
        '''
        if inputs is None: return None

        input_shape = inputs.size()
        assert len(input_shape) >= 2
        inputs = inputs.view(-1, input_shape[-1])

        running_mean = (self.running_sum / self.count)
        running_std = torch.clamp((self.running_sumsq / self.count \
                                  - running_mean.pow(2)).pow(0.5), min=self.eps)
        normed = torch.clamp((inputs - running_mean) / running_std, -5.0, 5.0)
        normed = normed.view(input_shape)
        return normed

    def running_mean(self):
        '''
            returning the running obseravtion mean for Tensorplex logging
            Returns:
                numpy array of current running observation mean
        '''
        running_mean = self.running_sum / self.count
        return running_mean.cpu().numpy()

    def running_std(self):
        '''
            returning the running standard deviation for Tensorplex Logging
            Returns:
                numpy array of running standard deviation
        '''
        running_std = ((self.running_sumsq / self.count) 
                     - (self.running_sum / self.count).pow(2)).pow(0.5)
        return running_std.cpu().numpy()
        
    def running_square(self):
        '''
            returning the running square mean for Tensorplex Logging
            Returns:
                running square mean
        '''
        running_square = self.running_sumsq / self.count
        return running_square.cpu().numpy()



"""
Serializes numpy and JSON-like objects
"""
import pickle
import base64
import hashlib
import json
import pyarrow as pa


def pa_serialize(obj):
    return pa.serialize(obj).to_buffer()


def pa_deserialize(binary):
    return pa.deserialize(binary)


_SERIALIZER = pa_serialize
_DESERIALIZER = pa_deserialize

# _SERIALIZER = pickle.dumps
# _DESERIALIZER = pickle.loads


def set_global_serializer(serializer, deserializer):
    """
    Call at the start of a script
    """
    assert callable(serializer) and callable(deserializer)
    global _SERIALIZER, _DESERIALIZER
    _SERIALIZER = serializer
    _DESERIALIZER = deserializer


def serialize(obj):
    """
    We can improve this function if we *really* need more memory efficiency
    """
    return _SERIALIZER(obj)


def deserialize(binary):
    """
    We can improve this function if we *really* need more memory efficiency
    """
    return _DESERIALIZER(binary)


def string_hash(s):
    assert isinstance(s, str)
    return binary_hash(s.encode('utf-8'))


def binary_hash(binary):
    """
    Low collision hash of any binary string
    For designating the 16-char object key in Redis.
    Runs at 200 mu-second per hash on Macbook pro.
    Only contains characters from [a-z][A-Z]+_
    """
    s = hashlib.md5(binary).digest()
    s = base64.b64encode(s)[:16]
    s = s.decode('utf-8')
    # return s.replace('/','_')
    return s


def pyobj_hash(obj):
    return binary_hash(serialize(obj))


def bytes2str(bytestring):
    if isinstance(bytestring, str):
        return bytestring
    else:
        return bytestring.decode('UTF-8')


def str2bytes(string):
    if isinstance(string, bytes):
        return string
    else:
        return string.encode('UTF-8')



import torch
import torchx as tx
import torchx.nn as nnx
import torch.nn as nn
import torchx.layers as L
import numpy as np

class CNNStemNetwork(nnx.Module):
    def __init__(self, D_obs, D_out, conv_channels=[16, 32], kernel_sizes=[8, 4], strides=[4,2]):
        super(CNNStemNetwork, self).__init__()
        layers = []
        for i in range(len(conv_channels)):
            layers.append(L.Conv2d(conv_channels[i], kernel_size=kernel_sizes[i], stride=strides[i]))
            layers.append(L.ReLU())
        layers.append(L.Flatten())
        layers.append(L.Linear(D_out))
        layers.append(L.ReLU())
        self.model = L.Sequential(*layers)

        # instantiate parameters
        self.model.build((None, *D_obs))

    def forward(self, obs):
        obs_shape = obs.size()
        if_high_dim = (len(obs_shape) == 5)
        if if_high_dim: # case of RNN input
            obs = obs.view(-1, *obs_shape[2:])  

        obs = self.model(obs)

        if if_high_dim:
            obs = obs.view(obs_shape[0], obs_shape[1], -1)
        return obs

class ActorNetworkX(nnx.Module):
    def __init__(self, D_in, D_act, hidden_sizes=[300, 200], use_layernorm=True):
        super(ActorNetworkX, self).__init__()

        xp_input = L.Placeholder((None, D_in))
        xp = L.Linear(hidden_sizes[0])(xp_input)
        xp = L.ReLU()(xp)
        if use_layernorm:
            # Normalize 1 dimension
            xp = L.LayerNorm(1)(xp)
        xp = L.Linear(hidden_sizes[1])(xp)
        xp = L.ReLU()(xp)
        if use_layernorm:
            xp = L.LayerNorm(1)(xp)
        xp = L.Linear(D_act)(xp)
        xp = L.Tanh()(xp)

        self.model = L.Functional(inputs=xp_input, outputs=xp)
        self.model.build((None, D_in))

    def forward(self, obs):
        return self.model(obs)

class CriticNetworkX(nnx.Module):
    def __init__(self, D_in, D_act, hidden_sizes=[400, 300], use_layernorm=True):
        super(CriticNetworkX, self).__init__()

        xp_input_obs = L.Placeholder((None, D_in))
        xp = L.Linear(hidden_sizes[0])(xp_input_obs)
        xp = L.ReLU()(xp)
        if use_layernorm:
            xp = L.LayerNorm(1)(xp)
        self.model_obs = L.Functional(inputs=xp_input_obs, outputs=xp)
        self.model_obs.build((None, D_in))

        xp_input_concat = L.Placeholder((None, hidden_sizes[0] + D_act))
        xp = L.Linear(hidden_sizes[1])(xp_input_concat)
        xp = L.ReLU()(xp)
        if use_layernorm:
            xp = L.LayerNorm(1)(xp)
        xp = L.Linear(1)(xp)

        self.model_concat = L.Functional(inputs=xp_input_concat, outputs=xp)
        self.model_concat.build((None, D_act + hidden_sizes[0]))

    def forward(self, obs, act):
        h_obs = self.model_obs(obs)
        h1 = torch.cat((h_obs, act), 1)
        value = self.model_concat(h1)
        return value

class PPO_ActorNetwork(nnx.Module):
    '''
        PPO custom actor network structure
    '''
    def __init__(self, D_obs, D_act, hidden_sizes=[64, 64], init_log_sig=0):
        '''
            Constructor for PPO actor network
            Args: 
                D_obs: observation space dimension, scalar
                D_act: action space dimension, scalar
                hidden_sizes: list of fully connected dimension
                init_log_sig: initial value for log standard deviation parameter
        '''
        super(PPO_ActorNetwork, self).__init__()
        # assumes D_obs here is the correct RNN hidden dim
        xp_input = L.Placeholder((None, D_obs))
        xp = L.Linear(hidden_sizes[0])(xp_input)
        xp = L.ReLU()(xp)
        xp = L.Linear(hidden_sizes[1])(xp)
        xp = L.ReLU()(xp)
        xp = L.Linear(D_act)(xp)
        xp = L.Tanh()(xp)

        self.model = L.Functional(inputs=xp_input, outputs=xp)
        self.model.build((None, D_obs))

        self.log_var = nn.Parameter(torch.zeros(1, D_act) + init_log_sig)

    def forward(self, obs):
        '''
            Forward pass of actor network. Input assume to be output of CNN
            and/or LSTM feature extractor
            Args:
                obs: batched tensor denotes the states
        '''
        obs_shape = obs.size()
        if_high_dim = (len(obs_shape) == 3)
        if if_high_dim: 
            obs = obs.view(-1, obs_shape[2])

        mean = self.model(obs)
        std  = torch.exp(self.log_var) * torch.ones(mean.size())

        action = torch.cat((mean, std), dim=1)
        if if_high_dim:
            action = action.view(obs_shape[0], obs_shape[1], -1)
        return action


class PPO_CriticNetwork(nnx.Module):
    '''
        PPO custom critic network structure
    '''
    def __init__(self, D_obs, hidden_sizes=[64, 64]):
        '''
            Constructor for PPO critic network
            Args: 
                D_obs: observation space dimension, scalar
                hidden_sizes: list of fully connected dimension
        '''
        super(PPO_CriticNetwork, self).__init__()
        # assumes D_obs here is the correct RNN hidden dim if necessary

        xp_input = L.Placeholder((None, D_obs))
        xp = L.Linear(hidden_sizes[0])(xp_input)
        xp = L.ReLU()(xp)
        xp = L.Linear(hidden_sizes[1])(xp)
        xp = L.ReLU()(xp)
        xp = L.Linear(1)(xp)

        self.model = L.Functional(inputs=xp_input, outputs=xp)
        self.model.build((None, D_obs))

    def forward(self, obs):
        '''
            Forward pass of actor network. Input assume to be output of CNN
            and/or LSTM feature extractor
            Args:
                obs: batched tensor denotes the states
        '''
        obs_shape = obs.size()
        if_high_dim = (len(obs_shape) == 3)
        if if_high_dim: 
            obs = obs.view(-1, obs_shape[2])

        v = self.model(obs)

        if if_high_dim:
            v = v.view(obs_shape[0], obs_shape[1], 1)
        return v




import cv2
import gym
import numpy as np

from collections import deque
from gym import spaces


class NoopResetEnv(gym.Wrapper):
    def __init__(self, env=None, noop_max=30):
        """Sample initial states by taking random number of no-ops on reset.
        No-op is assumed to be action 0.
        """
        super(NoopResetEnv, self).__init__(env)
        self.noop_max = noop_max
        self.override_num_noops = None
        assert env.unwrapped.get_action_meanings()[0] == 'NOOP'

    def _reset(self):
        """ Do no-op action for a number of steps in [1, noop_max]."""
        self.env.reset()
        if self.override_num_noops is not None:
            noops = self.override_num_noops
        else:
            noops = np.random.randint(1, self.noop_max + 1)
        assert noops > 0
        obs = None
        for _ in range(noops):
            obs, _, done, _ = self.env.step(0)
            if done:
                obs = self.env.reset()
        return obs


class FireResetEnv(gym.Wrapper):
    def __init__(self, env=None):
        """For environments where the user need to press FIRE for the game to start."""
        super(FireResetEnv, self).__init__(env)
        assert env.unwrapped.get_action_meanings()[1] == 'FIRE'
        assert len(env.unwrapped.get_action_meanings()) >= 3

    def _reset(self):
        self.env.reset()
        obs, _, done, _ = self.env.step(1)
        if done:
            self.env.reset()
        obs, _, done, _ = self.env.step(2)
        if done:
            self.env.reset()
        return obs


class EpisodicLifeEnv(gym.Wrapper):
    def __init__(self, env=None):
        """Make end-of-life == end-of-episode, but only reset on true game over.
        Done by DeepMind for the DQN and co. since it helps value estimation.
        """
        super(EpisodicLifeEnv, self).__init__(env)
        self.lives = 0
        self.was_real_done = True
        self.was_real_reset = False

    def _step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.was_real_done = done
        # check current lives, make loss of life terminal,
        # then update lives to handle bonus lives
        lives = self.env.unwrapped.ale.lives()
        if lives < self.lives and lives > 0:
            # for Qbert somtimes we stay in lives == 0 condtion for a few frames
            # so its important to keep lives > 0, so that we only reset once
            # the environment advertises done.
            done = True
        self.lives = lives
        return obs, reward, done, info

    def _reset(self):
        """Reset only when lives are exhausted.
        This way all states are still reachable even though lives are episodic,
        and the learner need not know about any of this behind-the-scenes.
        """
        if self.was_real_done:
            obs = self.env.reset()
            self.was_real_reset = True
        else:
            # no-op step to advance from terminal/lost life state
            obs, _, _, _ = self.env.step(0)
            self.was_real_reset = False
        self.lives = self.env.unwrapped.ale.lives()
        return obs


class MaxAndSkipEnv(gym.Wrapper):
    def __init__(self, env=None, skip=4):
        """Return only every `skip`-th frame"""
        super(MaxAndSkipEnv, self).__init__(env)
        # most recent raw observations (for max pooling across time steps)
        self._obs_buffer = deque(maxlen=2)
        self._skip = skip

    def _step(self, action):
        total_reward = 0.0
        done = None
        for _ in range(self._skip):
            obs, reward, done, info = self.env.step(action)
            self._obs_buffer.append(obs)
            total_reward += reward
            if done:
                break

        max_frame = np.max(np.stack(self._obs_buffer), axis=0)

        return max_frame, total_reward, done, info

    def _reset(self):
        """Clear past frame buffer and init. to first new_obs. from inner env."""
        self._obs_buffer.clear()
        obs = self.env.reset()
        self._obs_buffer.append(obs)
        return obs


class ProcessFrame84(gym.ObservationWrapper):
    def __init__(self, env, mode='pytorch'):
        super(ProcessFrame84, self).__init__(env)
        self.observation_space = spaces.Box(low=0, high=255, shape=(84, 84, 1))
        self.mode = mode

    def _observation(self, obs):
        return self.process(obs)

    def process(self, frame):
        if frame.size == 210 * 160 * 3:
            img = np.reshape(frame, [210, 160, 3]).astype(np.float32)
        elif frame.size == 250 * 160 * 3:
            img = np.reshape(frame, [250, 160, 3]).astype(np.float32)
        else:
            assert False, "Unknown resolution."
        img = img[:, :, 0] * 0.299 + img[:, :, 1] * 0.587 + img[:, :, 2] * 0.114
        resized_screen = cv2.resize(img, (84, 110), interpolation=cv2.INTER_AREA)
        x_t = resized_screen[18:102, :]
        if self.mode == 'pytorch':
            new_shape = [1, 84, 84]
        else:
            new_shape = [84, 84, 1]
        x_t = np.reshape(x_t, new_shape)
        return x_t.astype(np.uint8)


class ClippedRewardsWrapper(gym.RewardWrapper):
    def _reward(self, reward):
        """Change all the positive rewards to 1, negative to -1 and keep zero."""
        return np.sign(reward)


class LazyFrames(object):
    def __init__(self, frames, mode='pytorch'):
        """This object ensures that common frames between the observations are only stored once.
        It exists purely to optimize memory usage which can be huge for DQN's 1M frames replay
        buffers.

        This object should only be converted to numpy array before being passed to the model.

        You'd not belive how complex the previous solution was."""
        self._frames = frames
        self.mode = mode

    def __array__(self, dtype=None):
        axis = 0 if self.mode == 'pytorch' else 2
        out = np.concatenate(self._frames, axis=axis)
        if dtype is not None:
            out = out.astype(dtype)
        return out


class FrameStack(gym.Wrapper):
    def __init__(self, env, k, mode='pytorch'):
        """Stack k last frames.

        Returns lazy array, which is much more memory efficient.

        See Also
        --------
        baselines.common.atari_wrappers.LazyFrames
        """
        gym.Wrapper.__init__(self, env)
        self.k = k
        self.frames = deque([], maxlen=k)
        shp = env.observation_space.shape
        self.observation_space = spaces.Box(low=0, high=255, shape=(shp[0], shp[1], shp[2] * k))

    def _reset(self):
        ob = self.env.reset()
        for _ in range(self.k):
            self.frames.append(ob)
        return self._get_ob()

    def _step(self, action):
        ob, reward, done, info = self.env.step(action)
        self.frames.append(ob)
        return self._get_ob(), reward, done, info

    def _get_ob(self):
        assert len(self.frames) == self.k
        return LazyFrames(list(self.frames))


class ScaledFloatFrame(gym.ObservationWrapper):
    def _observation(self, obs):
        # careful! This undoes the memory optimization, use
        # with smaller replay buffers only.
        return np.array(obs).astype(np.float32) / 255.0


def wrap_dqn(env):
    """Apply a common set of wrappers for Atari games."""
    assert 'NoFrameskip' in env.spec.id
    env = EpisodicLifeEnv(env)
    env = NoopResetEnv(env, noop_max=30)
    env = MaxAndSkipEnv(env, skip=4)
    if 'FIRE' in env.unwrapped.get_action_meanings():
        env = FireResetEnv(env)
    env = ProcessFrame84(env)
    env = FrameStack(env, 4)
    env = ClippedRewardsWrapper(env)
    return env


class A2cProcessFrame(gym.Wrapper):
    def __init__(self, env):
        gym.Wrapper.__init__(self, env)
        self.observation_space = spaces.Box(low=0, high=255, shape=(84, 84, 1))

    def _step(self, action):
        ob, reward, done, info = self.env.step(action)
        return A2cProcessFrame.process(ob), reward, done, info

    def _reset(self):
        return A2cProcessFrame.process(self.env.reset())

    @staticmethod
    def process(frame):
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
        frame = cv2.resize(frame, (84, 84), interpolation=cv2.INTER_AREA)
        return frame.reshape(84, 84, 1)



import torch
import torch.nn as nn
import torch.nn.functional as F


class HeadBuilder(nn.Module):

    def __init__(self,
                 output_spec):

        super(HeadBuilder, self).__init__()
        self._dims = output_spec.dim
        self._type = output_spec.type
        self._head = None

    def forward(self, obs):

        in_dim = obs.size(1)
        out_dim = self._dims[0]

        if self._head is None:

            if self._type == 'discrete':
                self._head = nn.Linear(in_dim, out_dim)

            elif self._type == 'continuous':
                self._head = nn.Linear(in_dim, out_dim)

            elif self._type == 'gaussian':
                self._head = {
                    'mean': nn.Linear(in_dim, out_dim),
                    'std': nn.Linear(in_dim, out_dim)
                }

            elif self._type == 'scalar':
                self._head = nn.Linear(in_dim, 1)

            elif self._type == 'distributional':
                raise NotImplementedError

            else:
                raise ValueError('Unknown head type: {}'.format(type))

        if isinstance(self._head, dict):
            out = dict()
            for k, m in self._head.items():
                out[k] = m(obs)
        else:
            out = self._head(obs)

        return out

